# querytext.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Display a game selection rule.
"""

import tkinter

from solentware_misc.workarounds import dialogues

from .chessexception import ChessException
from .constants import (
    START_SELECTION_RULE_MARK,
    )
from ..core.querystatement import QueryStatement
from .eventspec import EventSpec
from .displayitems import DisplayItemsStub
from .gamerow import make_ChessDBrowGame
from ..core.chessrecord import ChessDBrecordGameTags
    

class QueryText(ChessException):

    """Game selection rule widget built from Text and Scrollbar widgets.
    """

    # True means selection rule can be edited
    _is_query_editable = False

    def __init__(
        self,
        panel,
        ui=None,
        items_manager=None,
        itemgrid=None,
        **ka):
        """Create widgets to display game selection rule.

        """
        super(QueryText, self).__init__(**ka)
        self.ui = ui

        # May be worth using a Null() instance for these two attributes.
        if items_manager is None:
            items_manager = DisplayItemsStub()
        self.items = items_manager
        self.itemgrid = itemgrid

        self.panel = panel
        self.score = tkinter.Text(
            master=self.panel,
            width=0,
            height=0,
            takefocus=tkinter.FALSE,
            undo=True,
            wrap=tkinter.WORD)
        self.scrollbar = tkinter.Scrollbar(
            master=self.panel,
            orient=tkinter.VERTICAL,
            takefocus=tkinter.FALSE,
            command=self.score.yview)
        self.score.configure(yscrollcommand=self.scrollbar.set)

        # Keyboard actions do nothing by default.
        self.disable_keyboard()

        # The popup menus for the selection rule

        self.viewmode_popup = tkinter.Menu(master=self.score, tearoff=False)
        self.viewmode_database_popup = tkinter.Menu(
            master=self.viewmode_popup, tearoff=False)
        self.inactive_popup = None
        self.viewmode_navigation_popup = None

        # Selection rule parser instance to process text.
        self.query_statement = QueryStatement()
        if ui.base_games.datasource:
            self.query_statement.dbset = ui.base_games.datasource.dbset

    def add_navigation_to_viewmode_popup(self, **kwargs):
        '''Add 'Navigation' entry to popup if not already present.'''

        # Cannot see a way of asking 'Does entry exist?' other than:
        try:
            self.viewmode_popup.index('Navigation')
        except:
            self.viewmode_navigation_popup = tkinter.Menu(
                master=self.viewmode_popup, tearoff=False)
            self.viewmode_popup.add_cascade(
                label='Navigation', menu=self.viewmode_navigation_popup)
            self.bind_navigation_for_viewmode_popup(**kwargs)
        
    def bind_for_viewmode(self):
        """Set keyboard bindings for game selection rule display."""

    def bind_navigation_for_viewmode_popup(self, bindings=None, order=None):
        """Set popup bindings for toplevel navigation."""
        if order is None:
            order = ()
        if bindings is None:
            bindings = {}
        for label, accelerator in order:
            function = bindings.get(label)
            if function is not None:
                self.viewmode_navigation_popup.add_command(
                    label=label,
                    command=self.try_command(
                        function, self.viewmode_navigation_popup),
                    accelerator=accelerator)

    def bind_score_pointer_for_board_navigation(self, switch):
        """Set or unset pointer bindings for game selection rule navigation."""
        ste = self.try_event
        for sequence, function in (
            ('<Control-ButtonPress-1>', ''),
            ('<ButtonPress-1>', ''),
            ('<ButtonPress-3>', ste(self.popup_viewmode_menu)),
            ):
            self.score.bind(sequence, '' if not switch else function)

    def give_focus_to_widget(self, event=None):
        """Do nothing and return 'break'.  Override in subclasses as needed."""
        return 'break'
        
    def set_query_statement(self, reset_undo=False):
        """Display the game selection rule as text.
        
        reset_undo causes the undo redo stack to be cleared if True.  Set True
        on first display of a selection rule for editing so that repeated
        Ctrl-Z in text editing mode recovers the original selection rule.
        
        """
        if not self._is_query_editable:
            self.score.configure(state=tkinter.NORMAL)
        self.score.delete('1.0', tkinter.END)
        self.map_query_statement()
        if not self._is_query_editable:
            self.score.configure(state=tkinter.DISABLED)
        if reset_undo:
            self.score.edit_reset()
        
    def set_statusbar_text(self):
        """Set status bar to display game selection rule name."""
        self.ui.statusbar.set_status_text(self.query_statement.get_name_text())

    def get_name_query_statement_text(self):
        """"""
        text = self.score.get('1.0', tkinter.END).strip()
        return text

    def map_query_statement(self):
        """"""
        # No mapping of tokens to text in widget (yet).
        self.score.insert(tkinter.INSERT,
                          self.query_statement.get_name_query_statement_text())
        
    def popup_inactive_menu(self, event=None):
        """Show the popup menu for a game selection rule in an inactive item.

        Subclasses may have particular entries to be the default which is
        implemented by overriding this method.

        """
        self.inactive_popup.tk_popup(*self.score.winfo_pointerxy())
        
    def popup_viewmode_menu(self, event=None):
        """Show the popup menu for game selection rule actions.

        Subclasses define particular entries for this menu.  This class adds
        no items to the menu.

        """
        if self.items.is_mapped_panel(self.panel):
            if self is not self.items.active_item:
                return 'break'
        self.viewmode_popup.tk_popup(*self.score.winfo_pointerxy())

    def refresh_game_list(self):
        """Display games matching game selection rule, empty on errors."""
        grid = self.itemgrid
        if grid is None:
            return
        if grid.get_database() is None:
            return
        self.ui.base_games.set_data_source(
            self.ui.selectionds(
                grid.datasource.dbhome,
                self.ui.base_games.datasource.dbset,
                self.ui.base_games.datasource.dbset,
                make_ChessDBrowGame(self.ui)),
            self.ui.base_games.on_data_change)
        qs = self.query_statement
        if qs.where_error:
            self.ui.base_games.datasource.get_selection_rule_games(None)
            self.ui.base_games.load_new_index()
            dialogues.showerror(
                'Display Game Selection Rule',
                qs.where_error.get_error_report(grid.datasource),
                )
        elif qs.where:
            qs.where.evaluate(
                grid.datasource.dbhome.record_finder(
                    grid.datasource.dbset,
                    ChessDBrecordGameTags))

            # Workaround problem with query ''.  See Where.evaluate() also.
            r = qs.where.node.get_root().result
            if r is None:
                self.ui.base_games.datasource.get_selection_rule_games(None)
            else:
                self.ui.base_games.datasource.get_selection_rule_games(
                    r.answer)
            self.ui.base_games.load_new_index()

        elif qs.get_query_statement_text():
            self.ui.base_games.load_new_index()
        #else:
        #    dialogues.showinfo(
        #        'Display Game Selection Rule',
        #        ''.join(('Game list not changed because active query ',
        #                 'has not yet been evaluated.',
        #                 )))

        # Get rid of the 'Please wait ...' status text.
        self.ui.statusbar.set_status_text()

    def disable_keyboard(self):
        """"""
        self.score.bind(EventSpec.selectiontext_disable_keypress[0],
                        lambda e: 'break')



# cqldisplay.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Widgets to display and edit Chess Query Language (ChessQL) statements.

These four classes display ChessQL statements in the main window: they are used
in the querygrid module.

The CQLDisplay class binds events to navigate between widgets.

The DatabaseCQLDisplay class adds delete record to the CQLDisplay class.

The DatabaseCQLInsert class adds insert record to the DatabaseCQLDisplay
class but does not bind delete record to any events.

The DatabaseCQLEdit class adds edit record to the DatabaseCQLInsert class
but does not bind delete record to any events.

These three classes display games in their own Toplevel widget: they are used
in the gamedbdelete, gamedbedit, and gamedbshow, modules.

The CQLDialogue class binds events to navigate between widgets.

The DialogueCQLDisplay class adds insert and delete record to the
CQLDialogue class.

The DialogueCQLEdit class adds insert and edit record to the CQLDialogue
class.

"""

import tkinter
import tkinter.messagebox

from solentware_misc.workarounds import dialogues

from solentware_grid.gui.dataedit import RecordEdit
from solentware_grid.gui.datadelete import RecordDelete
from solentware_grid.core.dataclient import DataNotify

from .cql import CQL
from .cqledit import CQLEdit
from ..core.chessrecord import ChessDBrecordPartial
from ..core.cqlstatement import CQLStatement
from .chessexception import ChessException
from .eventspec import EventSpec
from .display import Display
from .. import APPLICATION_NAME


class CQLDisplay(ChessException, Display):
    
    """Manage UI interaction with database for displayed ChessQL statements.

    CQLDisplay is a subclass of DataNotify so that modifications to
    the database record outside an instance prevent database update using
    the instance.  This class provides methods to update the database from
    the instance; and to switch to other displayed ChessQL statements.

    Subclasses provide the widget to display the ChessQL statements.
    
    """

    binding_labels = tuple(
        [b[1:] for b in (
            EventSpec.partialdisplay_to_position_grid,
            EventSpec.partialdisplay_to_active_game,
            EventSpec.partialdisplay_to_game_grid,
            EventSpec.partialdisplay_to_repertoire_grid,
            EventSpec.partialdisplay_to_active_repertoire,
            EventSpec.partialdisplay_to_repertoire_game_grid,
            EventSpec.partialdisplay_to_partial_grid,
            EventSpec.partialdisplay_to_previous_partial,
            EventSpec.partialdisplay_to_next_partial,
            EventSpec.partialdisplay_to_partial_game_grid,
            EventSpec.partialdisplay_to_selection_rule_grid,
            EventSpec.partialdisplay_to_active_selection_rule,
            EventSpec.tab_traverse_backward,
            EventSpec.tab_traverse_forward,
            )])

    def __init__(self, sourceobject=None, **ka):
        """Extend and link ChessQL statement to database.

        sourceobject - link to database.

        """
        super().__init__(**ka)
        self.blockchange = False
        if self.ui.base_partials.datasource:
            self.set_data_source(self.ui.base_partials.get_data_source())
        self.sourceobject = sourceobject
        self.insertonly = sourceobject is None
        self.recalculate_after_edit = sourceobject

    def _bind_for_board_navigation(self):
        """Set bindings to navigate ChessQL statement on pointer click."""
        self.bind_score_pointer_for_board_navigation(True)

    def bind_for_widget_navigation(self):
        """Set bindings to give focus to ChessQL statement on pointer click.
        """
        self.bind_score_pointer_for_widget_navigation(True)

    def bind_off(self):
        """Disable all bindings."""
        
        # Replicate structure of __bind_on for deleting bindings.
        for sequence, function in (
            (EventSpec.partialdisplay_to_partial_grid, ''),
            (EventSpec.partialdisplay_to_previous_partial, ''),
            (EventSpec.partialdisplay_to_next_partial, ''),
            (EventSpec.partialdisplay_to_partial_game_grid, ''),
            (EventSpec.partialdisplay_to_repertoire_grid, ''),
            (EventSpec.partialdisplay_to_active_repertoire, ''),
            (EventSpec.partialdisplay_to_repertoire_game_grid, ''),
            (EventSpec.partialdisplay_to_position_grid, ''),
            (EventSpec.partialdisplay_to_active_game, ''),
            (EventSpec.partialdisplay_to_selection_rule_grid, ''),
            (EventSpec.partialdisplay_to_active_selection_rule, ''),
            (EventSpec.partialdisplay_to_game_grid, ''),
            (EventSpec.export_from_partialdisplay, ''),
            ):
            if function:
                function = self.try_event(function)
            self.score.bind(sequence[0], function)

    # This method exists because the bindings cannot be done until all __init__
    # methods in superclasses have been called.
    def initialize_bindings(self):
        """Initialize all bindings."""
        self.__bind_on()
        for sequence, function in (
            (EventSpec.tab_traverse_forward,
             self.traverse_forward),
            (EventSpec.tab_traverse_backward,
             self.traverse_backward),
            (EventSpec.tab_traverse_round,
             self.traverse_round),
            ):
            if function:
                function = self.try_event(function)
            self.score.bind(sequence[0], function)

    def bind_on(self):
        """Enable all bindings."""
        self.__bind_on()

    def __bind_on(self):
        """Enable all bindings."""

        # Same bindings in initialize_bindings() and bind_on() in this class.
        for sequence, function in (
            ('<Control-ButtonPress-1>', ''),
            ('<ButtonPress-1>', self.try_event(self.give_focus_to_widget)),
            ('<ButtonPress-3>', self.try_event(self.popup_inactive_menu)),
            ):
            self.score.bind(sequence, function)
        for sequence, function in (
            (EventSpec.partialdisplay_to_partial_grid,
             self.set_focus_partial_grid),
            (EventSpec.partialdisplay_to_previous_partial,
             self.prior_item),
            (EventSpec.partialdisplay_to_next_partial,
             self.next_item),
            (EventSpec.partialdisplay_to_partial_game_grid,
             self.set_focus_partial_game_grid),
            (EventSpec.partialdisplay_to_repertoire_grid,
             self.set_focus_repertoire_grid),
            (EventSpec.partialdisplay_to_active_repertoire,
             self.set_focus_repertoirepanel_item),
            (EventSpec.partialdisplay_to_repertoire_game_grid,
             self.set_focus_repertoire_game_grid),
            (EventSpec.partialdisplay_to_position_grid,
             self.set_focus_position_grid),
            (EventSpec.partialdisplay_to_active_game,
             self.set_focus_gamepanel_item),
            (EventSpec.partialdisplay_to_selection_rule_grid,
             self.set_focus_selection_rule_grid),
            (EventSpec.partialdisplay_to_active_selection_rule,
             self.set_focus_selectionpanel_item),
            (EventSpec.partialdisplay_to_game_grid,
             self.set_focus_game_grid),
            (EventSpec.export_from_partialdisplay,
             self.export_partial),
            ):
            if function:
                function = self.try_event(function)
            self.score.bind(sequence[0], function)

    def _cycle_item(self, prior=False):
        """Select next ChessQL statement on display."""
        items = self.ui.partial_items
        losefocus = items.active_item
        losefocus.bind_for_widget_navigation()
        items.cycle_active_item(prior=prior)
        self.ui.configure_partial_grid()
        gainfocus = items.active_item
        gainfocus.refresh_game_list()
        gainfocus._bind_for_board_navigation()
        gainfocus.takefocus_widget.focus_set()
        gainfocus.set_statusbar_text()

    def give_focus_to_widget(self, event=None):
        """Select ChessQL statement on display by mouse click."""
        self.ui.set_bindings_on_item_losing_focus_by_pointer_click()
        losefocus, gainfocus = self.ui.partial_items.give_focus_to_widget(
            event.widget)
        if losefocus is not gainfocus:
            self.ui.configure_partial_grid()
            self.score.after(
                0,
                func=self.try_command(self.ui._set_partial_name, self.score))
            self.score.after(
                0,
                func=self.try_command(gainfocus.refresh_game_list, self.score))
        return 'break'

    def delete_item_view(self, event=None):
        """Remove ChessQL statement item from screen."""
        self.ui.delete_position_view(self)

    def insert_cql_statement_database(self, event=None):
        """Add ChessQL statement to database."""
        if self.ui.partial_items.active_item is None:
            dialogues.showinfo(
                title='Insert ChessQL Statement',
                message='No active ChessQL statement to insert into database.')
            return

        # This should see if ChessQL statement with same name already exists,
        # after checking for database open, and offer option to insert anyway.
        if self.ui.database is None:
            dialogues.showinfo(
                title='Insert ChessQL Statement',
                message='Cannot add ChessQL statement:\n\nNo database open.')
            return

        datasource = self.ui.base_partials.get_data_source()
        if datasource is None:
            dialogues.showinfo(
                title='Insert ChessQL Statement',
                message=''.join(('Cannot add ChessQL statement:\n\n',
                                 'Partial position list hidden.')))
            return
        updater = ChessDBrecordPartial()
        updater.value.process_statement(
            self.get_name_cql_statement_text())
        title = 'Insert ChessQL Statement'
        tname = title.replace('Insert ', '').replace('S', 's')
        if not len(updater.value.get_name_text()):
            dialogues.showerror(
                title=title,
                message=''.join((
                    "The '",
                    tname,
                    " has no name.\n\nPlease enter it's ",
                    "name as the first line of text.'")))
            return
        message=[
            ''.join((
                'Confirm request to add ',
                tname,
                ' named:\n\n',
                updater.value.get_name_text(),
                '\n\nto database.\n\n'))]
        if not updater.value.cql_error:
            if tkinter.messagebox.YES != dialogues.askquestion(
                title=title,
                message=''.join(message)):
                dialogues.showinfo(
                    title=title,
                    message=tname.join(('Add ',
                                        ' to database abandonned.')))
                return
        else:
            message.append(updater.value.cql_error.get_error_report())
            if tkinter.messagebox.YES != dialogues.askquestion(
                title=title,
                message=''.join(message)):
                dialogues.showinfo(
                    title=title,
                    message=tname.join(('Add ',
                                        ' to database abandonned.')))
                return
        editor = RecordEdit(updater, None)
        editor.set_data_source(datasource, editor.on_data_change)
        updater.set_database(editor.get_data_source().dbhome)
        updater.key.recno = None#0
        editor.put()
        dialogues.showinfo(
            title=title,
            message=''.join((tname,
                             ' "',
                             updater.value.get_name_text(),
                             '" added to database.')))
        
    def next_item(self, event=None):
        """Select next ChessQL statement display.

        Call _next_cql_statement after 1 millisecond to allow message display

        """
        if self.ui.partial_items.count_items_in_stack() > 1:
            self.ui._set_find_partial_name_games(0)
            self.score.after(
                1, func=self.try_command(self._next_cql_statement, self.score))

    def _next_cql_statement(self):
        """Generate next ChessQL statement display.

        Call from next_item only.

        """
        self._cycle_item(prior=False)

    def on_game_change(self, instance):
        """Recalculate list of games for ChessQL statement after game update.

        instance is ignored: it is assumed a recalculation is needed.

        """
        if self.sourceobject is not None:
            self._get_cql_statement_games_to_grid(self.cql_statement)#.match)

    def on_partial_change(self, instance):
        """Prevent update from self if instance refers to same record."""
        if instance.newrecord:

            # Editing an existing record.
            value = instance.newrecord.value
            key = instance.newrecord.key

        else:

            # Inserting a new record or deleting an existing record.
            value = instance.value
            key = instance.key

        if self.sourceobject is not None:
            if (key == self.sourceobject.key and
                self.datasource.dbname == self.sourceobject.dbname and
                self.datasource.dbset == self.sourceobject.dbset):
                self.blockchange = True

        pds = self.ui.partial_games.datasource
        if self.sourceobject is not None and key != self.sourceobject.key:
            pass
        elif self is not self.ui.partial_items.active_item:
            pass
        elif instance.newrecord is None:
            pds.forget_cql_statement_games(instance)
        elif instance.newrecord is False:
            try:
                pds.update_cql_statement_games(instance)
            except AttributeError as exc:
                if str(exc) == "'NoneType' object has no attribute 'answer'":
                    msg = ''.join(
                        ("Unable to add ChessQL statement to database, ",
                         "probably because an 'empty square' is in the query ",
                         "(eg '.a2-3'):\n\nThe reported  error is:\n\n",
                         str(exc),
                         ))
                else:
                    msg = ''.join(
                        ("Unable to add ChessQL statement to database:\n\n",
                         "The reported error is:\n\n",
                         str(exc),
                         ))
                dialogues.showinfo(
                    title='Insert ChessQL Statement',
                    message=msg)
                return
            except Exception as exc:
                msg = ''.join(
                    ("Unable to add ChessQL statement to database:\n\n",
                     "The reported error is:\n\n",
                     str(exc),
                     ))
                dialogues.showinfo(
                    title='Insert ChessQL Statement',
                    message=msg)
                return
        else:

            # Unfortunatly the existing list will have to be recalculated if
            # one of the caught exceptions occurs.
            pds.forget_cql_statement_games(instance)
            try:
                pds.update_cql_statement_games(instance.newrecord)
            except AttributeError as exc:
                if str(exc) == "'NoneType' object has no attribute 'answer'":
                    msg = ''.join(
                        ("Unable to edit ChessQL statement on database, ",
                         "probably because an 'empty square' is in the query ",
                         "(eg '.a2-3'):\n\nThe reported  error is:\n\n",
                         str(exc),
                         ))
                else:
                    msg = ''.join(
                        ("Unable to edit ChessQL statement on database:\n\n",
                         "The reported error is:\n\n",
                         str(exc),
                         ))
                dialogues.showinfo(
                    title='Insert ChessQL Statement',
                    message=msg)
                return
            except Exception as exc:
                msg = ''.join(
                    ("Unable to edit ChessQL statement on database:\n\n",
                     "The reported error is:\n\n",
                     str(exc),
                     ))
                dialogues.showinfo(
                    title='Insert ChessQL Statement',
                    message=msg)
                return

        if self is self.ui.partial_items.active_item:
            if self.sourceobject is not None and key == self.sourceobject.key:
                
                # Maybe should create a new CQLStatement instance.
                # self.cql_statement is a CQLStatement instance.
                # value is a ChessDBvaluePartial instance.
                self.cql_statement = value
                self.set_cql_statement()
                self._get_cql_statement_games_to_grid(value)

    def prior_item(self, event=None):
        """Select previous ChessQL statement display.

        Call _prior_cql_statement after 1 millisecond to allow message display

        """
        if self.ui.partial_items.count_items_in_stack() > 1:
            self.ui._set_find_partial_name_games(-2)
            self.score.after(
                1, func=self.try_command(
                    self._prior_cql_statement, self.score))

    def _prior_cql_statement(self):
        """Generate previous ChessQL statement display.

        Call from prior_item only.

        """
        self._cycle_item(prior=True)

    def set_insert_or_delete(self):
        """Convert edit display to insert display.

        ChessQL statements displayed for editing from a database are not closed
        if the database is closed.  They are converted to insert displays and
        can be used to add new records to databases opened later.
        
        """
        self.sourceobject = None

    def get_text_for_statusbar(self):
        """"""
        return ''.join(
            ('Please wait while finding games for ChessQL statement ',
             self.cql_statement.get_name_text(),
             ))

    def get_selection_text_for_statusbar(self):
        """"""
        return self.cql_statement.get_name_text()
        
    def bind_toplevel_navigation(self):
        """Set bindings for popup menu for CQLDisplay instance."""
        navigation_map = {
            EventSpec.partialdisplay_to_position_grid[1]:
            self.set_focus_position_grid,
            EventSpec.partialdisplay_to_active_game[1]:
            self.set_focus_gamepanel_item_command,
            EventSpec.partialdisplay_to_game_grid[1]:
            self.set_focus_game_grid,
            EventSpec.partialdisplay_to_repertoire_grid[1]:
            self.set_focus_repertoire_grid,
            EventSpec.partialdisplay_to_active_repertoire[1]:
            self.set_focus_repertoirepanel_item_command,
            EventSpec.partialdisplay_to_repertoire_game_grid[1]:
            self.set_focus_repertoire_game_grid,
            EventSpec.partialdisplay_to_partial_grid[1]:
            self.set_focus_partial_grid,
            EventSpec.partialdisplay_to_partial_game_grid[1]:
            self.set_focus_partial_game_grid,
            EventSpec.partialdisplay_to_selection_rule_grid[1]:
            self.set_focus_selection_rule_grid,
            EventSpec.partialdisplay_to_active_selection_rule[1]:
            self.set_focus_selectionpanel_item_command,
            EventSpec.tab_traverse_backward[1]:
            self.traverse_backward,
            EventSpec.tab_traverse_forward[1]:
            self.traverse_forward,
            }
        for nm, widget in (
            ({EventSpec.partialdisplay_to_previous_partial[1]: self.prior_item,
              EventSpec.partialdisplay_to_next_partial[1]: self.next_item,
              },
             self),
            ):
            nm.update(navigation_map)
            widget.add_navigation_to_viewmode_popup(
                bindings=nm, order=self.binding_labels)

    def set_game_list(self):
        """"""
        self.panel.after(
            0, func=self.try_command(self.ui._set_partial_name, self.panel))
        self.panel.after(
            0,
            func=self.try_command(self.refresh_game_list, self.panel))

    def traverse_backward(self, event=None):
        """Give focus to previous widget type in traversal order."""
        self.ui.give_focus_backward(self.ui.partial_items)
        return 'break'

    def traverse_forward(self, event=None):
        """Give focus to next widget type in traversal order."""
        self.ui.give_focus_forward(self.ui.partial_items)
        return 'break'

    def traverse_round(self, event=None):
        """Give focus to next widget within active item in traversal order."""
        return 'break'

    def _get_cql_statement_games_to_grid(self, statement):#match):
        """Populate Partial Position games grid with games selected by match.

        "match" is named for the CQL version-1.0 keyword which started a CQL
        statement.  Usage is "(match ..." which become "cql(" no later than
        version 5.0 of CQL.  Thus "cql" is now a better name for the argument.

        """
        p = self.ui.partial_games
        if len(p.keys):
            key = p.keys[0]
        else:
            key = None
        p.close_client_cursor()
        try:
            p.datasource.get_cql_statement_games(statement, self.sourceobject)
        except AttributeError as exc:
            if str(exc) == "'NoneType' object has no attribute 'answer'":
                msg = ''.join(
                    ("Unable to list games for ChessQL statement, ",
                     "probably because an 'empty square' is in the query ",
                     "(eg '.a2-3'):\n\nThe reported  error is:\n\n",
                     str(exc),
                     ))
            else:
                msg = ''.join(
                    ("Unable to list games for ChessQL statement:\n\n",
                     "The reported error is:\n\n",
                     str(exc),
                     ))
            dialogues.showinfo(
                title='Delete ChessQL Statement',
                message=msg)
        except Exception as exc:
            msg = ''.join(
                ("Unable to list games for ChessQL statement:\n\n",
                 "The reported error is:\n\n",
                 str(exc),
                 ))
            dialogues.showinfo(
                title='Delete ChessQL Statement',
                message=msg)
        p.fill_view(currentkey=key, exclude=False)
        if p.datasource.not_implemented:
            dialogues.showinfo(
                title='ChessQL Statement Not Implemented',
                message=''.join(('These filters are not implemented and ',
                                 'are ignored:\n\n',
                                 '\n'.join(sorted(
                                     grid.datasource.not_implemented)))))


class DatabaseCQLDisplay(CQLDisplay, CQL, DataNotify):
    
    """Display ChessQL statement from database and allow delete and insert.

    CQL provides the widget and CQLDisplay the database interface.
    
    """

    def __init__(self, **ka):
        """Create display widget and display ChessQL statement from database.

        See superclasses for argument descriptions.

        """
        super().__init__(**ka)
        self.initialize_bindings()

    def bind_off(self):
        """Disable all bindings."""
        super().bind_off()
        for sequence, function in (
            (EventSpec.databasepartialdisplay_insert, ''),
            (EventSpec.databasepartialdisplay_delete, ''),
            (EventSpec.databasepartialdisplay_dismiss, ''),
            ):
            if function:
                function = self.try_event(function)
            self.score.bind(sequence[0], function)

    # This method exists because the bindings cannot be done until all __init__
    # methods in superclasses have been called.
    def initialize_bindings(self):
        """Initialize all bindings."""
        super().initialize_bindings()

        # Here because superclass order is CQLDisplay, CQL.
        self.inactive_popup = tkinter.Menu(master=self.score, tearoff=False)

        for function, accelerator in (
            (self.set_focus_panel_item_command,
             EventSpec.databasepartialdisplay_make_active),
            (self.delete_item_view,
             EventSpec.databasepartialdisplay_dismiss_inactive),
            ):
            self.inactive_popup.add_command(
                label=accelerator[1],
                command=self.try_command(function, self.inactive_popup),
                accelerator=accelerator[2])
        self.__bind_on()
        for function, accelerator in (
            (self.delete_item_view,
             EventSpec.databasepartialdisplay_dismiss),
            ):
            self.viewmode_popup.add_command(
                label=accelerator[1],
                command=self.try_command(function, self.viewmode_popup),
                accelerator=accelerator[2])
        for function, accelerator in (
            (self.insert_cql_statement_database,
             EventSpec.databasepartialdisplay_insert),
            (self.delete_cql_statement_database,
             EventSpec.databasepartialdisplay_delete),
            ):
            self.viewmode_database_popup.add_command(
                label=accelerator[1],
                command=self.try_command(
                    function, self.viewmode_database_popup),
                accelerator=accelerator[2])
        self.bind_toplevel_navigation()

    def bind_on(self):
        """Enable all bindings."""
        super().bind_on()
        self.__bind_on()

    def __bind_on(self):
        """Common to bind_on() and initialize_bindings()"""
        for sequence, function in (
            (EventSpec.databasepartialdisplay_insert,
             self.insert_cql_statement_database),
            (EventSpec.databasepartialdisplay_delete,
             self.delete_cql_statement_database),
            (EventSpec.databasepartialdisplay_dismiss,
             self.delete_item_view),
            ):
            if function:
                function = self.try_event(function)
            self.score.bind(sequence[0], function)

    def delete_cql_statement_database(self, event=None):
        """Remove ChessQL statement from database."""
        if self.ui.database is None:
            dialogues.showinfo(
                title='Delete ChessQL Statement',
                message=''.join(('Cannot delete ChessQL statement:\n\n',
                                 'No database open.')))
            return
        datasource = self.ui.base_partials.get_data_source()
        if datasource is None:
            dialogues.showerror(
                title='Delete ChessQL Statement',
                message=''.join(('Cannot delete ChessQL statement:\n\n',
                                 'ChessQL statement list hidden.')))
            return
        if self.sourceobject is None:
            dialogues.showinfo(
                title='Delete ChessQL Statement',
                message=''.join(('The ChessQL statement to delete has not ',
                                 'been given.\n\nProbably because database ',
                                 'has been closed and opened since this copy ',
                                 'was displayed.')))
            return
        if self.blockchange:
            dialogues.showinfo(
                title='Delete ChessQL Statement',
                message='\n'.join((
                    'Cannot delete ChessQL statement.',
                    'Record has been amended since this copy displayed.')))
            return
        if tkinter.messagebox.YES != dialogues.askquestion(
            title='Delete ChessQL Statement',
            message='Confirm request to delete ChessQL statement.'):
            return
        s = self.cql_statement

        # Consider changing this since the call no longer ever returns None.
        if s.is_statement() is not None:
            
            v = self.sourceobject.value
            if (s.get_name_text() != v.get_name_text() or
                s.is_statement() != v.is_statement() or
                s.get_statement_text() != v.get_statement_text()):
                dialogues.showinfo(
                    title='Delete ChessQL Statement',
                    message='\n'.join((
                        'Cannot delete ChessQL statement.',
                        ' '.join((
                            'ChessQL statement on display is not same as',
                            'rule from record.')))))
                return
        editor = RecordDelete(self.sourceobject)
        editor.set_data_source(datasource, editor.on_data_change)
        editor.delete()
        dialogues.showinfo(
            title='Delete ChessQL Statement',
            message=''.join(('ChessQL statement "',
                             self.sourceobject.value.get_name_text(),
                             '" deleted from database.')))


class DatabaseCQLInsert(CQLDisplay, CQLEdit, DataNotify):
    
    """Display ChessQL statement from database allowing insert.

    CQLEdit provides the widget and CQLDisplay the database
    interface.
    
    """

    def __init__(self, **ka):
        """Create editor widget and display ChessQL statement from database.

        See superclasses for argument descriptions.

        """
        super().__init__(**ka)
        self.initialize_bindings()

    def bind_off(self):
        """Disable all bindings."""
        super().bind_off()
        for sequence, function in (
            (EventSpec.databasepartialedit_show_game_list, ''),
            (EventSpec.databasepartialedit_insert, ''),
            (EventSpec.databasepartialedit_dismiss, ''),
            ):
            if function:
                function = self.try_event(function)
            self.score.bind(sequence[0], function)

    # This method exists because the bindings cannot be done until all __init__
    # methods in superclasses have been called.
    def initialize_bindings(self):
        """Initialize all bindings."""
        super().initialize_bindings()

        # Here because superclass order is CQLDisplay, CQLEdit.
        self.inactive_popup = tkinter.Menu(master=self.score, tearoff=False)

        for function, accelerator in (
            (self.set_focus_panel_item_command,
             EventSpec.databasepartialedit_make_active),
            (self.delete_item_view,
             EventSpec.databasepartialedit_dismiss_inactive),
            ):
            self.inactive_popup.add_command(
                label=accelerator[1],
                command=self.try_command(function, self.inactive_popup),
                accelerator=accelerator[2])
        self.__bind_on()
        for function, accelerator in (
            (self.process_and_set_cql_statement_list,
             EventSpec.databasepartialedit_show_game_list),
            (self.delete_item_view,
             EventSpec.databasepartialedit_dismiss),
            ):
            self.viewmode_popup.add_command(
                label=accelerator[1],
                command=self.try_command(function, self.viewmode_popup),
                accelerator=accelerator[2])
        for function, accelerator in (
            (self.insert_cql_statement_database,
             EventSpec.databasepartialedit_insert),
            ):
            self.viewmode_database_popup.add_command(
                label=accelerator[1],
                command=self.try_command(
                    function, self.viewmode_database_popup),
                accelerator=accelerator[2])
        self.bind_toplevel_navigation()

    def bind_on(self):
        """Enable all bindings."""
        super().bind_on()
        self.__bind_on()

    def __bind_on(self):
        """Common to bind_on() and initialize_bindings()"""
        for sequence, function in (
            (EventSpec.databasepartialedit_show_game_list,
             self.process_and_set_cql_statement_list),
            (EventSpec.databasepartialedit_insert,
             self.insert_cql_statement_database),
            (EventSpec.databasepartialedit_dismiss,
             self.delete_item_view),
            ):
            if function:
                function = self.try_event(function)
            self.score.bind(sequence[0], function)

    def process_and_set_cql_statement_list(self, event=None):
        """Display games with position matching edited ChessQL statement."""
        s = CQLStatement()
        # Not sure this is needed or wanted.
        #s.dbset = self.ui.base_games.datasource.dbset
        s.process_statement(
            self.get_name_cql_statement_text())
        self.cql_statement = s
        try:
            self.refresh_game_list()
        except AttributeError as exc:
            if str(exc) == "'NoneType' object has no attribute 'answer'":
                msg = ''.join(
                    ("Unable to list games for ChessQL statement, probably ",
                     "because an 'empty square' is in the query ",
                     "(eg '.a2-3'):\n\nThe reported  error is:\n\n",
                     str(exc),
                     ))
            else:
                msg = ''.join(
                    ("Unable to list games for ChessQL statement:\n\n",
                     "The reported error is:\n\n",
                     str(exc),
                     ))
            dialogues.showinfo(
                title='ChessQL Statement',
                message=msg)
        except Exception as exc:
            msg = ''.join(
                ("Unable to list games for ChessQL statement:\n\n",
                 "The reported error is:\n\n",
                 str(exc),
                 ))
            dialogues.showinfo(
                title='ChessQL Statement',
                message=msg)
        return 'break'

    def insert_char_to_right(self, char):
        """"""
        r = super().insert_char_to_right(char)
        if r is None:
            return None
        elif r:
            return True
        else:
            self.recalculate_after_edit = None
            return None

    def delete_char_left(self, event):
        """"""
        if not self.current:
            return 'break'
        self.recalculate_after_edit = None
        return super().delete_char_left(event)

    def delete_char_right(self, event):
        """"""
        if not self.current:
            return 'break'
        self.recalculate_after_edit = None
        return super().delete_char_right(event)


class DatabaseCQLEdit(DatabaseCQLInsert):
    
    """Display ChessQL statement from database allowing edit and insert.

    DatabaseCQLInsert adds insert to the database interface.
    
    """

    def bind_off(self):
        """Disable all bindings."""
        super().bind_off()
        for sequence, function in (
            (EventSpec.databasepartialedit_update, ''),
            ):
            if function:
                function = self.try_event(function)
            self.score.bind(sequence[0], function)

    # This method exists because the bindings cannot be done until all __init__
    # methods in superclasses have been called.
    def initialize_bindings(self):
        """Initialize all bindings."""
        super().initialize_bindings()
        self.__bind_on()
        for function, accelerator in (
            (self.update_cql_statement_database,
             EventSpec.databasepartialedit_update),
            ):
            self.viewmode_database_popup.add_command(
                label=accelerator[1],
                command=self.try_command(
                    function, self.viewmode_database_popup),
                accelerator=accelerator[2])

    def bind_on(self):
        """Enable all bindings."""
        super().bind_on()
        self.__bind_on()

    def __bind_on(self):
        """Common to bind_on() and initialize_bindings()"""
        for sequence, function in (
            (EventSpec.databasepartialedit_update,
             self.update_cql_statement_database),
            ):
            if function:
                function = self.try_event(function)
            self.score.bind(sequence[0], function)

    def update_cql_statement_database(self, event=None):
        """Modify existing ChessQL statement record."""
        if self.ui.database is None:
            dialogues.showinfo(
                title='Edit ChessQL Statement',
                message='Cannot edit ChessQL statement:\n\nNo database open.')
            return
        datasource = self.ui.base_partials.get_data_source()
        if datasource is None:
            dialogues.showinfo(
                title='Edit ChessQL Statement',
                message=''.join(('Cannot edit ChessQL statement:\n\n',
                                 'Partial position list hidden.')))
            return
        if self.sourceobject is None:
            dialogues.showinfo(
                title='Edit ChessQL Statement',
                message=''.join(('The ChessQL statement to edit has not ',
                                 'been given.\n\nProbably because database ',
                                 'has been closed and opened since this copy ',
                                 'was displayed.')))
            return
        if self.blockchange:
            dialogues.showinfo(
                title='Edit ChessQL Statement',
                message='\n'.join((
                    'Cannot edit ChessQL statement.',
                    'It has been amended since this copy was displayed.')))
            return
        original = ChessDBrecordPartial()
        original.load_record(
            (self.sourceobject.key.recno,
             self.sourceobject.srvalue))

        # is it better to use DataClient directly?
        # Then original would not be used. Instead DataSource.new_row
        # gets record keyed by sourceobject and update is used to edit this.
        updater = ChessDBrecordPartial()
        updater.value.process_statement(
            self.get_name_cql_statement_text())
        title = 'Edit ChessQL Statement'
        tname = title.replace('Edit ', '').replace('S', 's')
        if not len(updater.value.get_name_text()):
            dialogues.showerror(
                title=title,
                message=''.join((
                    "The '",
                    tname,
                    " has no name.\n\nPlease enter it's ",
                    "name as the first line of text.'")))
            return
        message=[
            ''.join((
                'Confirm request to edit ',
                tname,
                ' named:\n\n',
                updater.value.get_name_text(),
                '\n\non database.\n\n'))]
        if not updater.value.cql_error:
            if tkinter.messagebox.YES != dialogues.askquestion(
                title=title,
                message=''.join(message)):
                dialogues.showinfo(
                    title=title,
                    message=tname.join(('Edit ',
                                        ' on database abandonned.')))
                return
        else:
            message.append(updater.value.cql_error.get_error_report())
            if tkinter.messagebox.YES != dialogues.askquestion(
                title=title,
                message=''.join(message)):
                dialogues.showinfo(
                    title=title,
                    message=tname.join(('Edit ',
                                        ' on database abandonned.')))
                return
        editor = RecordEdit(updater, original)
        editor.set_data_source(datasource, editor.on_data_change)
        updater.set_database(editor.get_data_source().dbhome)
        original.set_database(editor.get_data_source().dbhome)
        updater.key.recno = original.key.recno
        editor.edit()
        dialogues.showinfo(
            title=title,
            message=''.join((tname,
                             ' "',
                             updater.value.get_name_text(),
                             '" amended on database.')))


class CQLDialogue(ChessException):
    
    """Manage UI interaction with database for a displayed ChessQL statement.

    Subclasses provide the widget to display the ChessQL statement.
    
    """

    # Formally the same as GameDialogue in gamedisplay module.
    # Follow whatever style of assignment is used there if this binding_labels
    # needs to be non-empty.
    binding_labels = tuple()

    def initialize_bindings(self):
        """Enable all bindings."""
        self.bind_score_pointer_for_board_navigation(True)
        self.bind_toplevel_navigation()
        
    def bind_toplevel_navigation(self):
        """Set bindings for popup menu for CQLDialogue instance."""

        # Formally the same as GameDialogue in gamedisplay module.
        # The effect is add no bindings to an empty bindings map.
        navigation_map = {}
        for nm, widget in (
            ({},
             self),
            ):
            nm.update(navigation_map)
            widget.add_navigation_to_viewmode_popup(
                bindings=nm, order=self.binding_labels)


class DialogueCQLDisplay(CQL, CQLDialogue):
    
    """Display a ChessQL statement from a database allowing delete and insert.
    """

    def __init__(self, **ka):
        """Extend and link ChessQL statement to database.
        """
        super().__init__(**ka)
        self.initialize_bindings()


class DialogueCQLEdit(CQLEdit, CQLDialogue):
    
    """Display a ChessQL statement from a database allowing edit and insert.
    """

    def __init__(self, **ka):
        """Extend and link ChessQL statement to database.

        """
        super().__init__(**ka)
        self.initialize_bindings()



# create_chesstab_database.py
# Copyright 2020 Roger Marsh
# Licence: See LICENSE.txt (BSD licence)

"""Create empty ChessTab database with chosen database engine and segment size.
"""

from solentware_base.tools import create_database

try:
    from ..unqlite import chessunqlite
except ImportError: # Not ModuleNotFoundError for Pythons earlier than 3.6
    chessunqlite = None
try:
    from ..vedis import chessvedis
except ImportError: # Not ModuleNotFoundError for Pythons earlier than 3.6
    chessvedis = None
if create_database._deny_sqlite3:
    chesssqlite3 = None
else:
    try:
        from ..sqlite import chesssqlite3
    except ImportError: # Not ModuleNotFoundError for Pythons earlier than 3.6
        chesssqlite3 = None
try:
    from ..apsw import chessapsw
except ImportError: # Not ModuleNotFoundError for Pythons earlier than 3.6
    chessapsw = None
try:
    from ..db import chessdb
except ImportError: # Not ModuleNotFoundError for Pythons earlier than 3.6
    chessdb = None
try:
    from ..dpt import chessdpt
except ImportError: # Not ModuleNotFoundError for Pythons earlier than 3.6
    chessdpt = None
try:
    from ..ndbm import chessndbm
except ImportError: # Not ModuleNotFoundError for Pythons earlier than 3.6
    chessndbm = None
try:
    from ..gnu import chessgnu
except ImportError: # Not ModuleNotFoundError for Pythons earlier than 3.6
    chessgnu = None


class CreateChessTabDatabase(create_database.CreateDatabase):
    """Create a ChessTab database."""

    _START_TEXT = ''.join(
        ('ChessTab would create a new database with the top-left engine, ',
         'and segment size 4000.',
         ))

    def __init__(self):
        """Build the user interface."""
        engines = {}
        if chessunqlite:
            engines[chessunqlite.unqlite_database.unqlite
                    ] = chessunqlite.ChessDatabase
        if chessvedis:
            engines[chessvedis.vedis_database.vedis] = chessvedis.ChessDatabase
        if chesssqlite3:
            engines[chesssqlite3.sqlite3_database.sqlite3
                    ] = chesssqlite3.ChessDatabase
        if chessapsw:
            engines[chessapsw.apsw_database.apsw] = chessapsw.ChessDatabase
        if chessdb:
            engines[chessdb.bsddb3_database.bsddb3] = chessdb.ChessDatabase
        if chessdpt:
            engines[chessdpt.dpt_database.dptapi] = chessdpt.ChessDatabase
        if chessndbm:
            engines[chessndbm.ndbm_database.dbm.ndbm] = chessndbm.ChessDatabase
        if chessgnu:
            engines[chessgnu.gnu_database.dbm.gnu] = chessgnu.ChessDatabase
        super().__init__(title='Create ChessTab Database', engines=engines)


if __name__ == '__main__':

    CreateChessTabDatabase().root.mainloop()



# gamelistgrid.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Grids for listing details of games on chess database.
"""

import tkinter

from solentware_grid.datagrid import DataGrid

from pgn_read.core.constants import REPERTOIRE_TAG_ORDER, UNKNOWN_RESULT

from ..core.chessrecord import (
    ChessDBrecordGameUpdate,
    ChessDBrecordRepertoireUpdate,
    PLAYER_NAME_TAGS,
    re_normalize_player_name,
    )
from .chessexception import ChessException
from .gamerow import ChessDBrowGame
from .positionrow import ChessDBrowPosition
from .repertoirerow import ChessDBrowRepertoire
from .gamedisplay import DatabaseGameDisplay, DatabaseGameEdit
from .repertoiredisplay import DatabaseRepertoireDisplay, DatabaseRepertoireEdit
from .constants import (
    EMPTY_SEVEN_TAG_ROSTER,
    STATUS_SEVEN_TAG_ROSTER_EVENT,
    STATUS_SEVEN_TAG_ROSTER_SCORE,
    STATUS_SEVEN_TAG_ROSTER_PLAYERS,
    EMPTY_REPERTOIRE_GAME,
    )
from ..core.filespec import POSITIONS_FIELD_DEF
from ..core import exporters
from .eventspec import EventSpec, DummyEvent
from .display import Display


class GameListGrid(ChessException, DataGrid, Display):

    """A DataGrid for lists of chess games.

    Subclasses provide navigation and extra methods appropriate to list use.
    
    """

    def __init__(self, parent, ui):
        """Extend with link to user interface object.

        parent - see superclass
        ui - container for user interface widgets and methods.

        """
        super(GameListGrid, self).__init__(parent=parent)
        self.gcanvas.configure(takefocus=tkinter.FALSE)
        self.data.configure(takefocus=tkinter.FALSE)
        self.frame.configure(takefocus=tkinter.FALSE)
        self.hsbar.configure(takefocus=tkinter.FALSE)
        self.vsbar.configure(takefocus=tkinter.FALSE)
        self.ui = ui
        for sequence, function in (
            (EventSpec.tab_traverse_forward,
             self.traverse_forward),
            (EventSpec.tab_traverse_backward,
             self.traverse_backward),
            (EventSpec.tab_traverse_round,
             self.traverse_round),
            ):
            if function:
                function = self.try_event(function)
            self.frame.bind(sequence[0], function)

    def display_selected_item(self, key):
        """Create display and return a DatabaseGameDisplay for selected game."""
        selected = self.get_visible_record(key)
        if selected is None:
            return None
        # Should the Frame containing board and score be created here and
        # passed to DatabaseGameDisplay. (Which needs 'import Tkinter' above.)
        # Rather than passing the container where the Frame created by
        # DatabaseGameDisplay is to be put.
        # Yes because DatabaseGameEdit (see edit_selected_item) includes
        # extra widgets. Want to say game.widget.destroy() eventually.
        # Read make_display_widget for DatabaseGameDisplay and DatabaseGameEdit.
        game = self.make_display_widget(selected)
        self.ui.add_game_to_display(game)
        self.ui.game_items.increment_object_count(selected.srvalue)
        self.ui.game_items.set_itemmap(game, selected.srvalue)
        self.ui.set_properties_on_all_game_grids(key)
        return game

    def make_display_widget(self, sourceobject):
        """Return a DatabaseGameDisplay for sourceobject."""
        game = DatabaseGameDisplay(
            master=self.ui.view_games_pw,
            ui=self.ui,
            items_manager=self.ui.game_items,
            itemgrid=self.ui.game_games,
            sourceobject=sourceobject)
        game.set_position_analysis_data_source()
        game.pgn.get_first_game(sourceobject.get_srvalue())
        return game
        
    def edit_selected_item(self, key):
        """Create display and return a DatabaseGameEdit for selected game."""
        selected = self.get_visible_record(key)
        if selected is None:
            return None
        # Should the Frame containing board and score be created here and
        # passed to DatabaseGameDisplay. (Which needs 'import Tkinter' above.)
        # Rather than passing the container where the Frame created by
        # DatabaseGameEdit is to be put.
        # Yes because DatabaseGameDisplay (see display_selected_item) includes
        # less widgets. Want to say game.widget.destroy() eventually.
        # Read make_edit_widget for DatabaseGameDisplay and DatabaseGameEdit.
        game = self.make_edit_widget(selected)
        self.ui.add_game_to_display(game)
        self.ui.game_items.increment_object_count(selected.srvalue)
        self.ui.game_items.set_itemmap(game, selected.srvalue)
        self.ui.set_properties_on_all_game_grids(key)
        return game
        
    def make_edit_widget(self, sourceobject):
        """Return a DatabaseGameEdit for sourceobject."""
        game = DatabaseGameEdit(
            master=self.ui.view_games_pw,
            ui=self.ui,
            items_manager=self.ui.game_items,
            itemgrid=self.ui.game_games,
            sourceobject=sourceobject)
        game.set_position_analysis_data_source()
        game.pgn.get_first_game(sourceobject.get_srvalue())
        return game

    def set_properties(self, key, dodefaultaction=True):
        """Return True if properties for game key set or False."""
        if super(GameListGrid, self).set_properties(
            key, dodefaultaction=False):
            return True
        if self.ui.game_items.object_display_count(self.objects[key].srvalue):
            self.objects[key].set_background_on_display(
                self.get_row_widgets(key))
            self.set_row_under_pointer_background(key)
            return True
        if dodefaultaction:
            self.objects[key].set_background_normal(self.get_row_widgets(key))
            self.set_row_under_pointer_background(key)
            return True
        return False

    def set_row(self, key, dodefaultaction=True, **kargs):
        """Return row widget for game key or None."""
        row = super(GameListGrid, self).set_row(
            key, dodefaultaction=False, **kargs)
        if row is not None:
            return row
        if key not in self.keys:
            return None
        if self.ui.game_items.object_display_count(self.objects[key].srvalue):
            return self.objects[key].grid_row_on_display(**kargs)
        if dodefaultaction:
            return self.objects[key].grid_row_normal(**kargs)
        else:
            return None
        
    def select_down(self):
        """Extend to show selection summary in status bar."""
        super(GameListGrid, self).select_down()
        self.set_selection_text()
        
    def select_up(self):
        """Extend to show selection summary in status bar."""
        super(GameListGrid, self).select_up()
        self.set_selection_text()
        
    def cancel_selection(self):
        """Extend to clear selection summary from status bar."""
        if self.selection:
            self.ui.statusbar.set_status_text('')
        super(GameListGrid, self).cancel_selection()

    def launch_delete_record(self, key, modal=True):
        """Create delete dialogue."""
        oldobject = ChessDBrecordGameUpdate()
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue))
        self.create_delete_dialog(
            self.objects[key],
            oldobject,
            modal,
            title='Delete Game')

    def launch_edit_record(self, key, modal=True):
        """Create edit dialogue."""
        self.create_edit_dialog(
            self.objects[key],
            ChessDBrecordGameUpdate(),
            ChessDBrecordGameUpdate(),
            False,
            modal,
            title='Edit Game')

    def launch_edit_show_record(self, key, modal=True):
        """Create edit dialogue including reference copy of original."""
        self.create_edit_dialog(
            self.objects[key],
            ChessDBrecordGameUpdate(),
            ChessDBrecordGameUpdate(),
            True,
            modal,
            title='Edit Game')

    def launch_insert_new_record(self, modal=True):
        """Create insert dialogue."""
        newobject = ChessDBrecordGameUpdate()
        instance = self.datasource.new_row()
        instance.srvalue = repr(EMPTY_SEVEN_TAG_ROSTER + UNKNOWN_RESULT)
        self.create_edit_dialog(
            instance,
            newobject,
            None,
            False,
            modal,
            title='New Game')

    def launch_show_record(self, key, modal=True):
        """Create show dialogue."""
        oldobject = ChessDBrecordGameUpdate()
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue))
        self.create_show_dialog(
            self.objects[key],
            oldobject,
            modal,
            title='Show Game')
        
    def create_edit_dialog(
        self, instance, newobject, oldobject, showinitial, modal, title=''):
        """Extend to do chess initialization"""
        for x in (newobject, oldobject):
            if x:
                x.load_record((instance.key.pack(), instance.srvalue))
        super(GameListGrid, self).create_edit_dialog(
            instance, newobject, oldobject, showinitial, modal, title=title)

    def fill_view(
        self,
        currentkey=None,
        down=True,
        topstart=True,
        exclude=True,
        ):
        """Delegate to superclass if database is open otherwise do nothing."""

        # Intend to put this in superclass but must treat the DataClient objects
        # being scrolled as a database to do this properly.  Do this when these
        # objects have been given a database interface used when the database
        # is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of games
        # from PGN files; which can take many hours.
        
        if self.get_database() is not None:
            super(GameListGrid, self).fill_view(
                currentkey=currentkey,
                down=down,
                topstart=topstart,
                exclude=exclude,
                )

    def load_new_index(self):
        """Delegate to superclass if database is open otherwise do nothing."""

        # Intend to put this in superclass but must treat the DataClient objects
        # being scrolled as a database to do this properly.  Do this when these
        # objects have been given a database interface used when the database
        # is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of games
        # from PGN files; which can take many hours.
        
        if self.get_database() is not None:
            super(GameListGrid, self).load_new_index()

    def load_new_partial_key(self, key):
        """Delegate to superclass if database is open otherwise do nothing."""

        # Intend to put this in superclass but must treat the DataClient objects
        # being scrolled as a database to do this properly.  Do this when these
        # objects have been given a database interface used when the database
        # is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of games
        # from PGN files; which can take many hours.
        
        if self.get_database() is not None:
            super(GameListGrid, self).load_new_partial_key(key)

    def on_configure_canvas(self, event=None):
        """Delegate to superclass if database is open otherwise do nothing."""

        # Intend to put this in superclass but must treat the DataClient objects
        # being scrolled as a database to do this properly.  Do this when these
        # objects have been given a database interface used when the database
        # is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of games
        # from PGN files; which can take many hours.
        
        if self.get_database() is not None:
            super(GameListGrid, self).on_configure_canvas(event=event)

    def on_data_change(self, instance):
        """Delegate to superclass if database is open otherwise do nothing."""

        # Intend to put this in superclass but must treat the DataClient objects
        # being scrolled as a database to do this properly.  Do this when these
        # objects have been given a database interface used when the database
        # is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of games
        # from PGN files; which can take many hours.
        
        if self.get_database() is not None:
            super(GameListGrid, self).on_data_change(instance)

    def add_navigation_to_popup(self):
        '''Add 'Navigation' entry to popup menu if not already present.'''

        # Cannot see a way of asking 'Does entry exist?' other than:
        try:
            self.menupopup.index('Navigation')
        except:
            self.menupopup_navigation = tkinter.Menu(
                master=self.menupopup, tearoff=False)
            self.menupopup.add_cascade(
                label='Navigation', menu=self.menupopup_navigation)

    def add_navigation_to_popup_no_row(self):
        '''Add 'Navigation' entry to popup menu if not already present.'''

        # Cannot see a way of asking 'Does entry exist?' other than:
        try:
            self.menupopupnorow.index('Navigation')
        except:
            self.menupopup_navigation_no_row = tkinter.Menu(
                master=self.menupopupnorow, tearoff=False)
            self.menupopupnorow.add_cascade(
                label='Navigation', menu=self.menupopup_navigation_no_row)

    def traverse_backward(self, event=None):
        """Give focus to previous widget type in traversal order."""
        self.ui.give_focus_backward(self)
        return 'break'

    def traverse_forward(self, event=None):
        """Give focus to next widget type in traversal order."""
        self.ui.give_focus_forward(self)
        return 'break'

    def traverse_round(self, event=None):
        """Give focus to next widget within active item in traversal order."""
        return 'break'

    def set_focus(self):
        """Give focus to this widget."""
        self.frame.focus_set()
        if self.ui.single_view:
            self.ui.show_just_panedwindow_with_focus(self.frame)

    def is_payload_available(self):
        """Return True if grid is connected to a database."""
        ds = self.get_data_source()
        if ds is None:
            return False
        if ds.get_database() is None:

            # Avoid exception scrolling visible grid not connected to database.
            # Make still just be hack to cope with user interface activity
            # while importing data.
            self.clear_grid_keys()

            return False
        return True
        
    def set_move_highlight(self, game):
        """Set move highlight at current position in game.

        In particular a game displayed from the list of games matching a
        position is shown at that position.

        """
        if game != None:
            if game.current:
                game.set_move_highlight(game.current, True, True)

    def bookmark_down(self):
        """Extend to show selection summary in status bar."""
        super(GameListGrid, self).bookmark_down()
        self.set_selection_text()
        
    def bookmark_up(self):
        """Extend to show selection summary in status bar."""
        super(GameListGrid, self).bookmark_up()
        self.set_selection_text()

    def export_pgn(self, event=None):
        """Export selected games as PGN."""
        exporters.export_grid_games_as_pgn(
            self,
            self.ui.get_export_filename('Games', pgn=True))

    def export_archive_pgn(self, event=None):
        """Export selected games as PGN archive."""
        exporters.export_grid_games_as_archive_pgn(
            self,
            self.ui.get_export_filename('Archive Games', pgn=True))

    def export_rav_pgn(self, event=None):
        """Export selected games as PGN excluding all commentary tokens."""
        exporters.export_grid_games_as_rav_pgn(
            self,
            self.ui.get_export_filename('RAV Games', pgn=True))

    def focus_set_frame(self, event=None):
        """Adjust widget which is losing focus then delegate to superclass."""
        self.ui.set_bindings_on_item_losing_focus_by_pointer_click()
        super().focus_set_frame(event=event)

    def bind_for_widget_without_focus(self):
        """Return True if this item has the focus about to be lost."""
        if self.get_frame().focus_displayof() != self.get_frame():
            return False

        # Nothing to do on losing focus.
        return True
        
    def get_top_widget(self):
        """Return topmost widget for game display.

        The topmost widget is put in a container widget in some way

        """
        # Superclass DataGrid.get_frame() method returns the relevant widget.
        # Name, get_top_widget, is compatible with Game and Partial names.
        return self.get_frame()
        

# Because of possible changes to GameListGrid to support database update after
# introducing subclasses of PGN to do just the required work it may not be
# possible for PartialPositionGames to be subclass of GameListGrid.
class PartialPositionGames(GameListGrid):

    """Customized GameListGrid for list of games matching a partial position.

    The grid is populated by a ChessQueryLanguageDS instance from the dpt.cqlds
    or basecore.cqlds modules.
    """

    def __init__(self, ui):
        """Extend with partial position grid definition and bindings.

        ui - container for user interface widgets and methods.

        """
        super(PartialPositionGames, self).__init__(ui.position_partials_pw, ui)
        self.make_header(ChessDBrowGame.header_specification)
        self.__bind_on()
        for function, accelerator in (
            (self.display_game_from_popup,
             EventSpec.display_game_from_partial_game_grid),
            (self.edit_game_from_popup,
             EventSpec.edit_game_from_partial_game_grid),
            (self.export_archive_pgn,
             EventSpec.export_archive_pgn_from_partial_game_grid),
            (self.export_rav_pgn,
             EventSpec.export_rav_pgn_from_partial_game_grid),
            (self.export_pgn,
             EventSpec.export_pgn_from_partial_game_grid),
            ):
            self.menupopup.add_command(
                label=accelerator[1],
                command=self.try_command(function, self.menupopup),
                accelerator=accelerator[2])
        self.add_navigation_to_popup()
        self.add_navigation_to_popup_no_row()
        for function, accelerator in (
            (self.set_focus_position_grid,
             EventSpec.partial_game_grid_to_position_grid),
            (self.set_focus_gamepanel_item_command,
             EventSpec.partial_game_grid_to_active_game),
            (self.set_focus_game_grid,
             EventSpec.partial_game_grid_to_game_grid),
            (self.set_focus_repertoire_grid,
             EventSpec.partial_game_grid_to_repertoire_grid),
            (self.set_focus_repertoirepanel_item_command,
             EventSpec.partial_game_grid_to_active_repertoire),
            (self.set_focus_repertoire_game_grid,
             EventSpec.partial_game_grid_to_repertoire_game_grid),
            (self.set_focus_partial_grid,
             EventSpec.partial_game_grid_to_partial_grid),
            (self.set_focus_partialpanel_item_command,
             EventSpec.partial_game_grid_to_active_partial),
            (self.set_focus_selection_rule_grid,
             EventSpec.partial_game_grid_to_selection_rule_grid),
            (self.set_focus_selectionpanel_item_command,
             EventSpec.partial_game_grid_to_active_selection_rule),
            (self.traverse_backward,
             EventSpec.tab_traverse_backward),
            (self.traverse_forward,
             EventSpec.tab_traverse_forward),
            ):
            for m in (self.menupopup_navigation,
                      self.menupopup_navigation_no_row):
                m.add_command(
                    label=accelerator[1],
                    command=self.try_command(function, m),
                    accelerator=accelerator[2])

    def bind_off(self):
        """Disable all bindings."""
        super(PartialPositionGames, self).bind_off()
        for sequence, function in (
            (EventSpec.partial_game_grid_to_partial_grid, ''),
            (EventSpec.partial_game_grid_to_active_partial, ''),
            (EventSpec.partial_game_grid_to_repertoire_grid, ''),
            (EventSpec.partial_game_grid_to_active_repertoire, ''),
            (EventSpec.partial_game_grid_to_repertoire_game_grid, ''),
            (EventSpec.partial_game_grid_to_position_grid, ''),
            (EventSpec.partial_game_grid_to_active_game,
             self.set_focus_gamepanel_item),
            (EventSpec.partial_game_grid_to_game_grid, ''),
            (EventSpec.partial_game_grid_to_selection_rule_grid, ''),
            (EventSpec.partial_game_grid_to_active_selection_rule, ''),
            (EventSpec.display_game_from_partial_game_grid, ''),
            (EventSpec.edit_game_from_partial_game_grid, ''),
            (EventSpec.export_archive_pgn_from_partial_game_grid, ''),
            (EventSpec.export_rav_pgn_from_partial_game_grid, ''),
            (EventSpec.export_pgn_from_partial_game_grid, ''),
            ):
            if function:
                function = self.try_event(function)
            self.frame.bind(sequence[0], function)

    def bind_on(self):
        """Enable all bindings."""
        super(PartialPositionGames, self).bind_on()
        self.__bind_on()

    def __bind_on(self):
        """Enable all bindings."""
        for sequence, function in (
            (EventSpec.partial_game_grid_to_partial_grid,
             self.set_focus_partial_grid),
            (EventSpec.partial_game_grid_to_active_partial,
             self.set_focus_partialpanel_item),
            (EventSpec.partial_game_grid_to_repertoire_grid,
             self.set_focus_repertoire_grid),
            (EventSpec.partial_game_grid_to_active_repertoire,
             self.set_focus_repertoirepanel_item),
            (EventSpec.partial_game_grid_to_repertoire_game_grid,
             self.set_focus_repertoire_game_grid),
            (EventSpec.partial_game_grid_to_position_grid,
             self.set_focus_position_grid),
            (EventSpec.partial_game_grid_to_active_game,
             self.set_focus_gamepanel_item),
            (EventSpec.partial_game_grid_to_game_grid,
             self.set_focus_game_grid),
            (EventSpec.partial_game_grid_to_selection_rule_grid,
             self.set_focus_selection_rule_grid),
            (EventSpec.partial_game_grid_to_active_selection_rule,
             self.set_focus_selectionpanel_item),
            (EventSpec.display_game_from_partial_game_grid,
             self.display_game),
            (EventSpec.edit_game_from_partial_game_grid,
             self.edit_game),
            (EventSpec.export_archive_pgn_from_partial_game_grid,
             self.export_archive_pgn),
            (EventSpec.export_rav_pgn_from_partial_game_grid,
             self.export_rav_pgn),
            (EventSpec.export_pgn_from_partial_game_grid,
             self.export_pgn),
            ):
            if function:
                function = self.try_event(function)
            self.frame.bind(sequence[0], function)

    def display_game(self, event=None):
        """Display selected game and cancel selection."""
        self.set_move_highlight(
            self.display_selected_item(self.get_visible_selected_key()))
        self.cancel_selection()

    def display_game_from_popup(self, event=None):
        """Display game selected by pointer."""
        self.set_move_highlight(
            self.display_selected_item(self.pointer_popup_selection))

    def edit_game(self, event=None):
        """Display selected game with editing allowed and cancel selection."""
        self.set_move_highlight(
            self.edit_selected_item(self.get_visible_selected_key()))
        self.cancel_selection()

    def edit_game_from_popup(self, event=None):
        """Display game with editing allowed selected by pointer."""
        self.set_move_highlight(
            self.edit_selected_item(self.pointer_popup_selection))

    def on_game_change(self, instance):
        # datasource refers to a set derived from file and may need
        # to be recreated
        if self.get_data_source() is None:
            return
        super(PartialPositionGames, self).on_data_change(instance)

    def set_selection_text(self):
        """Set status bar to display main PGN Tags."""
        if self.selection:
            ss0 = self.selection[0]
            if ss0 in self.objects:
                t = self.objects[ss0].value.collected_game[1]
                self.ui.statusbar.set_status_text(
                    '  '.join([t.get(k, '')
                               for k in STATUS_SEVEN_TAG_ROSTER_PLAYERS]) +
                    self.ui.partial_items.active_item.get_selection_text_for_statusbar(
                        ).join(('   (', ')')))
        else:
            self.ui.statusbar.set_status_text('')

    def is_visible(self):
        """Return True if list of games matching partials is displayed."""
        return str(self.get_frame()) in self.ui.position_partials_pw.panes()

    def is_payload_available(self):
        """Return True if connected to database and games displayed."""
        if not super().is_payload_available():
            return False
        return self.ui.partial_items.is_visible()

    def make_display_widget(self, sourceobject):
        """Return a DatabaseGameDisplay for sourceobject."""
        game = super().make_display_widget(sourceobject)
        game.set_game()
        return game
        
    def make_edit_widget(self, sourceobject):
        """Return a DatabaseGameEdit for sourceobject."""
        game = super().make_edit_widget(sourceobject)
        game.set_game(reset_undo=True)
        return game

    def focus_set_frame(self, event=None):
        """Delegate to superclass then set toolbar widget states."""
        super().focus_set_frame(event=event)
        self.ui.set_toolbarframe_disabled()


class GamePositionGames(GameListGrid):

    """Customized GameListGrid for list of games matching a game position.

    The grid is populated by a FullPositionDS instance from the
    dpt.fullpositionds or basecore.fullpositionds modules.
    """

    def __init__(self, ui):
        """Extend with position grid definition and bindings.

        ui - container for user interface widgets and methods.

        """
        super(GamePositionGames, self).__init__(ui.position_games_pw, ui)
        self.make_header(ChessDBrowPosition.header_specification)
        self.__bind_on()
        for function, accelerator in (
            (self.display_game_from_popup,
             EventSpec.display_game_from_position_grid),
            (self.edit_game_from_popup,
             EventSpec.edit_game_from_position_grid),
            (self.export_archive_pgn,
             EventSpec.export_archive_pgn_from_position_grid),
            (self.export_rav_pgn,
             EventSpec.export_rav_pgn_from_position_grid),
            (self.export_pgn,
             EventSpec.export_pgn_from_position_grid),
            (self.up_one_transposition,
             EventSpec.cycle_up_one_repeat_in_position_grid),
            (self.down_one_transposition,
             EventSpec.cycle_down_one_repeat_in_position_grid),
            ):
            self.menupopup.add_command(
                label=accelerator[1],
                command=self.try_command(function, self.menupopup),
                accelerator=accelerator[2])
        self.add_navigation_to_popup()
        self.add_navigation_to_popup_no_row()
        for function, accelerator in (
            (self.set_focus_gamepanel_item_command,
             EventSpec.position_grid_to_active_game),
            (self.set_focus_game_grid,
             EventSpec.position_grid_to_game_grid),
            (self.set_focus_repertoire_grid,
             EventSpec.position_grid_to_repertoire_grid),
            (self.set_focus_repertoirepanel_item_command,
             EventSpec.position_grid_to_active_repertoire),
            (self.set_focus_repertoire_game_grid,
             EventSpec.position_grid_to_repertoire_game_grid),
            (self.set_focus_partial_grid,
             EventSpec.position_grid_to_partial_grid),
            (self.set_focus_partialpanel_item_command,
             EventSpec.position_grid_to_active_partial),
            (self.set_focus_partial_game_grid,
             EventSpec.position_grid_to_partial_game_grid),
            (self.set_focus_selection_rule_grid,
             EventSpec.position_grid_to_selection_rule_grid),
            (self.set_focus_selectionpanel_item_command,
             EventSpec.position_grid_to_active_selection_rule),
            (self.traverse_backward,
             EventSpec.tab_traverse_backward),
            (self.traverse_forward,
             EventSpec.tab_traverse_forward),
            ):
            for m in (self.menupopup_navigation,
                      self.menupopup_navigation_no_row):
                m.add_command(
                    label=accelerator[1],
                    command=self.try_command(function, m),
                    accelerator=accelerator[2])

    def bind_off(self):
        """Disable all bindings."""
        super(GamePositionGames, self).bind_off()
        for sequence, function in (
            (EventSpec.position_grid_to_repertoire_grid, ''),
            (EventSpec.position_grid_to_active_repertoire, ''),
            (EventSpec.position_grid_to_repertoire_game_grid, ''),
            (EventSpec.position_grid_to_partial_grid, ''),
            (EventSpec.position_grid_to_active_partial, ''),
            (EventSpec.position_grid_to_partial_game_grid, ''),
            (EventSpec.position_grid_to_active_game,
             self.set_focus_gamepanel_item),
            (EventSpec.position_grid_to_game_grid, ''),
            (EventSpec.position_grid_to_selection_rule_grid, ''),
            (EventSpec.position_grid_to_active_selection_rule, ''),
            (EventSpec.display_game_from_position_grid, ''),
            (EventSpec.edit_game_from_position_grid, ''),
            (EventSpec.cycle_up_one_repeat_in_position_grid, ''),
            (EventSpec.cycle_down_one_repeat_in_position_grid, ''),
            (EventSpec.export_archive_pgn_from_position_grid, ''),
            (EventSpec.export_rav_pgn_from_position_grid, ''),
            (EventSpec.export_pgn_from_position_grid, ''),
            ):
            if function:
                function = self.try_event(function)
            self.frame.bind(sequence[0], function)

    def bind_on(self):
        """Enable all bindings."""
        super(GamePositionGames, self).bind_on()
        self.__bind_on()

    def __bind_on(self):
        """Enable all bindings."""
        for sequence, function in (
            (EventSpec.position_grid_to_repertoire_grid,
             self.set_focus_repertoire_grid),
            (EventSpec.position_grid_to_active_repertoire,
             self.set_focus_repertoirepanel_item),
            (EventSpec.position_grid_to_repertoire_game_grid,
             self.set_focus_repertoire_game_grid),
            (EventSpec.position_grid_to_partial_grid,
             self.set_focus_partial_grid),
            (EventSpec.position_grid_to_active_partial,
             self.set_focus_partialpanel_item),
            (EventSpec.position_grid_to_partial_game_grid,
             self.set_focus_partial_game_grid),
            (EventSpec.position_grid_to_active_game,
             self.set_focus_gamepanel_item),
            (EventSpec.position_grid_to_game_grid,
             self.set_focus_game_grid),
            (EventSpec.position_grid_to_selection_rule_grid,
             self.set_focus_selection_rule_grid),
            (EventSpec.position_grid_to_active_selection_rule,
             self.set_focus_selectionpanel_item),
            (EventSpec.display_game_from_position_grid,
             self.display_game),
            (EventSpec.edit_game_from_position_grid,
             self.edit_game),
            (EventSpec.cycle_up_one_repeat_in_position_grid,
             self.up_one_transposition),
            (EventSpec.cycle_down_one_repeat_in_position_grid,
             self.down_one_transposition),
            (EventSpec.export_archive_pgn_from_position_grid,
             self.export_archive_pgn),
            (EventSpec.export_rav_pgn_from_position_grid,
             self.export_rav_pgn),
            (EventSpec.export_pgn_from_position_grid,
             self.export_pgn),
            ):
            if function:
                function = self.try_event(function)
            self.frame.bind(sequence[0], function)

    def display_game(self, event=None):
        """Display selected game and cancel selection."""
        self.set_move_highlight(
            self.display_selected_item(self.get_visible_selected_key()))
        self.cancel_selection()

    def display_game_from_popup(self, event=None):
        """Display game selected by pointer."""
        self.set_move_highlight(
            self.display_selected_item(self.pointer_popup_selection))

    def edit_game(self, event=None):
        """Display selected game with editing allowed and cancel selection."""
        self.set_move_highlight(
            self.edit_selected_item(self.get_visible_selected_key()))
        self.cancel_selection()

    def edit_game_from_popup(self, event=None):
        """Display game with editing allowed selected by pointer."""
        self.set_move_highlight(
            self.edit_selected_item(self.pointer_popup_selection))

    def set_row(self, key, dodefaultaction=True, **kargs):
        """Return row widget for game key or None.

        Add arguments to **kargs for grid_row method in PositionRow class.
        
        """
        kargs.update(
            position=self.datasource.fullposition,
            context=self.ui.get_active_game_move())
        return super(GamePositionGames, self).set_row(
            key, dodefaultaction=dodefaultaction, **kargs)

    def on_game_change(self, instance):
        # datasource refers to a set derived from file and may need
        # to be recreated
        if self.get_data_source() is None:
            return
        # It may be on_data_change(None) should prevent GamePositionGames
        # being cleared on deleting game, but it does not.
        super(GamePositionGames, self).on_data_change(instance)

    def down_one_transposition(self, event=None): # Hack for menu item
        """Cycle forward one tranposition to position in selected game."""
        if self.selection:
            ss0 = self.selection[0]
            if ss0 in self.objects:
                self.objects[ss0].score.down_one_transposition()
        
    def up_one_transposition(self, event=None): # Hack for menu item
        """Cycle backward one tranposition to position in selected game."""
        if self.selection:
            ss0 = self.selection[0]
            if ss0 in self.objects:
                self.objects[ss0].score.up_one_transposition()
        
    def set_selection_text(self):
        """Set status bar to display main PGN Tags."""
        if self.selection:
            ss0 = self.selection[0]
            if ss0 in self.objects:
                t = self.objects[ss0].value.collected_game[1]
                self.ui.statusbar.set_status_text(
                    '  '.join([t.get(k, '')
                               for k in STATUS_SEVEN_TAG_ROSTER_EVENT]))
        else:
            self.ui.statusbar.set_status_text('')
    
    def is_visible(self):
        """Return True if list of matching games is displayed."""
        return str(self.get_frame()) in self.ui.position_games_pw.panes()

    def is_payload_available(self):
        """Return True if connected to database and games displayed."""
        if not super().is_payload_available():
            return False
        return self.ui.game_items.is_visible()

    def make_display_widget(self, sourceobject):
        """Return a DatabaseGameDisplay for sourceobject."""
        game = super().make_display_widget(sourceobject)

        # decode_move_number may be put in self.game
        #game.set_game(
        #    str(sourceobject.decode_move_number(self.selection[0][-1])))
        game.set_game()

        return game
        
    def make_edit_widget(self, sourceobject):
        """Return a DatabaseGameEdit for sourceobject."""
        game = super().make_edit_widget(sourceobject)

        # decode_move_number may be put in self.game
        #game.set_game(
        #    str(sourceobject.decode_move_number(self.selection[0][-1])),
        #    reset_undo=True)
        game.set_game(reset_undo=True)
        
        return game

    def focus_set_frame(self, event=None):
        """Delegate to superclass then set toolbar widget states."""
        super().focus_set_frame(event=event)
        self.ui.set_toolbarframe_disabled()


class TagRosterGrid(GameListGrid):

    """Customized GameListGrid for list of games on database.

    The grid is usually populated by a DataSource instance from the
    solentware_grid.core.dataclient module, either all games or by index or
    filter, but can be populated by a ChessQLGames instance from the dpt.cqlds
    or basecore.cqlds modules, when a selection rule is invoked.
    """

    def __init__(self, ui):
        """Extend with definition and bindings for games on database grid.

        ui - container for user interface widgets and methods.

        """
        super(TagRosterGrid, self).__init__(ui.games_pw, ui)
        self.make_header(ChessDBrowGame.header_specification)
        self.__bind_on()
        for function, accelerator in (
            (self.display_game_from_popup,
             EventSpec.display_game_from_game_grid),
            (self.edit_game_from_popup,
             EventSpec.edit_game_from_game_grid),
            (self.export_archive_pgn,
             EventSpec.export_archive_pgn_from_game_grid),
            (self.export_rav_pgn,
             EventSpec.export_rav_pgn_from_game_grid),
            (self.export_pgn,
             EventSpec.export_pgn_from_game_grid),
            ):
            self.menupopup.add_command(
                label=accelerator[1],
                command=self.try_command(function, self.menupopup),
                accelerator=accelerator[2])
        self.add_navigation_to_popup()
        self.add_navigation_to_popup_no_row()
        for function, accelerator in (
            (self.set_focus_position_grid,
             EventSpec.game_grid_to_position_grid),
            (self.set_focus_gamepanel_item_command,
             EventSpec.game_grid_to_active_game),
            (self.set_focus_repertoire_grid,
             EventSpec.game_grid_to_repertoire_grid),
            (self.set_focus_repertoirepanel_item_command,
             EventSpec.game_grid_to_active_repertoire),
            (self.set_focus_repertoire_game_grid,
             EventSpec.game_grid_to_repertoire_game_grid),
            (self.set_focus_partial_grid,
             EventSpec.game_grid_to_partial_grid),
            (self.set_focus_partialpanel_item_command,
             EventSpec.game_grid_to_active_partial),
            (self.set_focus_partial_game_grid,
             EventSpec.game_grid_to_partial_game_grid),
            (self.set_focus_selection_rule_grid,
             EventSpec.game_grid_to_selection_rule_grid),
            (self.set_focus_selectionpanel_item_command,
             EventSpec.game_grid_to_active_selection_rule),
            (self.traverse_backward,
             EventSpec.tab_traverse_backward),
            (self.traverse_forward,
             EventSpec.tab_traverse_forward),
            ):
            for m in (self.menupopup_navigation,
                      self.menupopup_navigation_no_row):
                m.add_command(
                    label=accelerator[1],
                    command=self.try_command(function, m),
                    accelerator=accelerator[2])

    def bind_off(self):
        """Disable all bindings."""
        super(TagRosterGrid, self).bind_off()
        for sequence, function in (
            (EventSpec.game_grid_to_repertoire_grid, ''),
            (EventSpec.game_grid_to_active_repertoire, ''),
            (EventSpec.game_grid_to_repertoire_game_grid, ''),
            (EventSpec.game_grid_to_partial_grid, ''),
            (EventSpec.game_grid_to_active_partial, ''),
            (EventSpec.game_grid_to_partial_game_grid, ''),
            (EventSpec.game_grid_to_position_grid, ''),
            (EventSpec.game_grid_to_active_game,
             self.set_focus_gamepanel_item),
            (EventSpec.game_grid_to_selection_rule_grid, ''),
            (EventSpec.game_grid_to_active_selection_rule, ''),
            (EventSpec.display_game_from_game_grid, ''),
            (EventSpec.edit_game_from_game_grid, ''),
            (EventSpec.export_archive_pgn_from_game_grid, ''),
            (EventSpec.export_rav_pgn_from_game_grid, ''),
            (EventSpec.export_pgn_from_game_grid, ''),
            ):
            if function:
                function = self.try_event(function)
            self.frame.bind(sequence[0], function)

    def bind_on(self):
        """Enable all bindings."""
        super(TagRosterGrid, self).bind_on()
        self.__bind_on()

    def __bind_on(self):
        """Enable all bindings."""
        for sequence, function in (
            (EventSpec.game_grid_to_repertoire_grid,
             self.set_focus_repertoire_grid),
            (EventSpec.game_grid_to_active_repertoire,
             self.set_focus_repertoirepanel_item),
            (EventSpec.game_grid_to_repertoire_game_grid,
             self.set_focus_repertoire_game_grid),
            (EventSpec.game_grid_to_partial_grid,
             self.set_focus_partial_grid),
            (EventSpec.game_grid_to_active_partial,
             self.set_focus_partialpanel_item),
            (EventSpec.game_grid_to_partial_game_grid,
             self.set_focus_partial_game_grid),
            (EventSpec.game_grid_to_position_grid,
             self.set_focus_position_grid),
            (EventSpec.game_grid_to_active_game,
             self.set_focus_gamepanel_item),
            (EventSpec.game_grid_to_selection_rule_grid,
             self.set_focus_selection_rule_grid),
            (EventSpec.game_grid_to_active_selection_rule,
             self.set_focus_selectionpanel_item),
            (EventSpec.display_game_from_game_grid,
             self.display_game),
            (EventSpec.edit_game_from_game_grid,
             self.edit_game),
            (EventSpec.export_archive_pgn_from_game_grid,
             self.export_archive_pgn),
            (EventSpec.export_rav_pgn_from_game_grid,
             self.export_rav_pgn),
            (EventSpec.export_pgn_from_game_grid,
             self.export_pgn),
            ):
            if function:
                function = self.try_event(function)
            self.frame.bind(sequence[0], function)

    def display_game(self, event=None):
        """Display selected game and cancel selection."""
        self.display_selected_item(self.get_visible_selected_key())
        self.cancel_selection()

    def display_game_from_popup(self, event=None):
        """Display game selected by pointer."""
        self.display_selected_item(self.pointer_popup_selection)

    def edit_game(self, event=None):
        """Display selected game with editing allowed and cancel selection."""
        self.edit_selected_item(self.get_visible_selected_key())
        self.cancel_selection()

    def edit_game_from_popup(self, event=None):
        """Display game with editing allowed selected by pointer."""
        self.edit_selected_item(self.pointer_popup_selection)

    def on_game_change(self, instance):
        # may turn out to be just to catch datasource is None
        if self.get_data_source() is None:
            return
        super(TagRosterGrid, self).on_data_change(instance)

    def set_selection_text(self):
        """Set status bar to display main PGN Tags."""
        if self.selection:
            ss0 = self.selection[0]
            if ss0 in self.objects:
                tags = self.objects[ss0].value.collected_game[1]
                self.ui.statusbar.set_status_text(
                    '  '.join(
                        [tags.get(k, '')
                         for k in STATUS_SEVEN_TAG_ROSTER_SCORE] +
                        [' '.join(
                            [t.group()
                             for t in self.objects[ss0].value.collected_game[2]
                             if t])]))
        else:
            self.ui.statusbar.set_status_text('')
    
    def is_visible(self):
        """Return True if list of games is displayed."""
        return str(self.get_frame()) in self.ui.games_pw.panes()

    def make_display_widget(self, sourceobject):
        """Return a DatabaseGameDisplay for sourceobject."""
        game = super().make_display_widget(sourceobject)
        game.set_game()
        return game
        
    def make_edit_widget(self, sourceobject):
        """Return a DatabaseGameEdit for sourceobject."""
        game = super().make_edit_widget(sourceobject)
        game.set_game(reset_undo=True)
        return game

    def focus_set_frame(self, event=None):
        """Delegate to superclass then set toolbar widget states."""
        super().focus_set_frame(event=event)
        ui = self.ui
        if ui.base_games.datasource.dbname in ui.allow_filter:
            ui.set_toolbarframe_normal(ui.move_to_game, ui.filter_game)
        else:
            ui.set_toolbarframe_disabled()

    def set_selection(self, key):
        """Hack to fix edge case when inserting records using apsw or sqlite3.
        
        Workaround a KeyError exception when a record is inserted while a grid
        keyed by a secondary index with only one key value in the index is on
        display.
        
        """
        try:
            super().set_selection(key)
        except KeyError:
            tkinter.messagebox.showinfo(
                title='Insert Game Workaround',
                message=''.join(
                    ('All records have same name on this display.\n\nThe new ',
                     'record has been inserted but you need to switch to ',
                     'another index, and back, to see the record in the list.',
                     )))

    def move_to_row_in_grid(self, key):
        """Navigate grid to nearest row starting with key"""
        if self.datasource.dbname in PLAYER_NAME_TAGS:
            if isinstance(key, str):
                key = ' '.join(re_normalize_player_name.findall(key))
        super().move_to_row_in_grid(key)

    def load_new_partial_key(self, key):
        """Transform key if it's a str and a player's name then delegate."""
        if self.datasource.dbname in PLAYER_NAME_TAGS:
            if isinstance(key, str):
                key = ' '.join(re_normalize_player_name.findall(key))
        super().load_new_partial_key(key)


class RepertoireGrid(GameListGrid):

    """Customized GameListGrid for list of repertoires on database.
    """

    def __init__(self, ui):
        """Extend with definition and bindings for games on database grid.

        ui - container for user interface widgets and methods.

        """
        super(RepertoireGrid, self).__init__(ui.repertoires_pw, ui)
        self.make_header(ChessDBrowRepertoire.header_specification)
        self.__bind_on()
        for function, accelerator in (
            (self.display_game_from_popup,
             EventSpec.display_game_from_repertoire_grid),
            (self.edit_game_from_popup,
             EventSpec.edit_game_from_repertoire_grid),
            (self.export_rav_pgn,
             EventSpec.export_rav_pgn_from_repertoire_grid),
            (self.export_pgn,
             EventSpec.export_pgn_from_repertoire_grid),
            ):
            self.menupopup.add_command(
                label=accelerator[1],
                command=self.try_command(function, self.menupopup),
                accelerator=accelerator[2])
        self.add_navigation_to_popup()
        self.add_navigation_to_popup_no_row()
        for function, accelerator in (
            (self.set_focus_position_grid,
             EventSpec.repertoire_grid_to_position_grid),
            (self.set_focus_gamepanel_item_command,
             EventSpec.repertoire_grid_to_active_game),
            (self.set_focus_game_grid,
             EventSpec.repertoire_grid_to_game_grid),
            (self.set_focus_repertoirepanel_item_command,
             EventSpec.repertoire_grid_to_active_repertoire),
            (self.set_focus_repertoire_game_grid,
             EventSpec.repertoire_grid_to_repertoire_game_grid),
            (self.set_focus_partial_grid,
             EventSpec.repertoire_grid_to_partial_grid),
            (self.set_focus_partialpanel_item_command,
             EventSpec.repertoire_grid_to_active_partial),
            (self.set_focus_partial_game_grid,
             EventSpec.repertoire_grid_to_partial_game_grid),
            (self.set_focus_selection_rule_grid,
             EventSpec.repertoire_grid_to_selection_rule_grid),
            (self.set_focus_selectionpanel_item_command,
             EventSpec.repertoire_grid_to_active_selection_rule),
            (self.traverse_backward,
             EventSpec.tab_traverse_backward),
            (self.traverse_forward,
             EventSpec.tab_traverse_forward),
            ):
            for m in (self.menupopup_navigation,
                      self.menupopup_navigation_no_row):
                m.add_command(
                    label=accelerator[1],
                    command=self.try_command(function, m),
                    accelerator=accelerator[2])

    def bind_off(self):
        """Disable all bindings."""
        super(RepertoireGrid, self).bind_off()
        for sequence, function in (
            (EventSpec.repertoire_grid_to_active_repertoire, ''),
            (EventSpec.repertoire_grid_to_repertoire_game_grid, ''),
            (EventSpec.repertoire_grid_to_partial_grid, ''),
            (EventSpec.repertoire_grid_to_active_partial, ''),
            (EventSpec.repertoire_grid_to_partial_game_grid, ''),
            (EventSpec.repertoire_grid_to_position_grid, ''),
            (EventSpec.repertoire_grid_to_active_game,
             self.set_focus_gamepanel_item),
            (EventSpec.repertoire_grid_to_game_grid, ''),
            (EventSpec.repertoire_grid_to_selection_rule_grid, ''),
            (EventSpec.repertoire_grid_to_active_selection_rule, ''),
            (EventSpec.display_game_from_repertoire_grid, ''),
            (EventSpec.edit_game_from_repertoire_grid, ''),
            (EventSpec.export_rav_pgn_from_repertoire_grid, ''),
            (EventSpec.export_pgn_from_repertoire_grid, ''),
            ):
            if function:
                function = self.try_event(function)
            self.frame.bind(sequence[0], function)

    def bind_on(self):
        """Enable all bindings."""
        super(RepertoireGrid, self).bind_on()
        self.__bind_on()

    def __bind_on(self):
        """Enable all bindings."""
        for sequence, function in (
            (EventSpec.repertoire_grid_to_active_repertoire,
             self.set_focus_repertoirepanel_item),
            (EventSpec.repertoire_grid_to_repertoire_game_grid,
             self.set_focus_repertoire_game_grid),
            (EventSpec.repertoire_grid_to_partial_grid,
             self.set_focus_partial_grid),
            (EventSpec.repertoire_grid_to_active_partial,
             self.set_focus_partialpanel_item),
            (EventSpec.repertoire_grid_to_partial_game_grid,
             self.set_focus_partial_game_grid),
            (EventSpec.repertoire_grid_to_position_grid,
             self.set_focus_position_grid),
            (EventSpec.repertoire_grid_to_active_game,
             self.set_focus_gamepanel_item),
            (EventSpec.repertoire_grid_to_game_grid,
             self.set_focus_game_grid),
            (EventSpec.repertoire_grid_to_selection_rule_grid,
             self.set_focus_selection_rule_grid),
            (EventSpec.repertoire_grid_to_active_selection_rule,
             self.set_focus_selectionpanel_item),
            (EventSpec.display_game_from_repertoire_grid,
             self.display_game),
            (EventSpec.edit_game_from_repertoire_grid,
             self.edit_game),
            (EventSpec.export_rav_pgn_from_repertoire_grid,
             self.export_rav_pgn),
            (EventSpec.export_pgn_from_repertoire_grid,
             self.export_pgn),
            ):
            if function:
                function = self.try_event(function)
            self.frame.bind(sequence[0], function)

    def display_game(self, event=None):
        """Display selected repertoire and cancel selection."""
        self.display_selected_item(self.get_visible_selected_key())
        self.cancel_selection()

    def display_game_from_popup(self, event=None):
        """Display repertoire selected by pointer."""
        self.display_selected_item(self.pointer_popup_selection)

    def edit_game(self, event=None):
        """Display selected repertoire for editing and cancel selection."""
        self.edit_selected_item(self.get_visible_selected_key())
        self.cancel_selection()

    def edit_game_from_popup(self, event=None):
        """Display repertoire with editing allowed selected by pointer."""
        self.edit_selected_item(self.pointer_popup_selection)

    def on_game_change(self, instance):
        # may turn out to be just to catch datasource is None
        if self.get_data_source() is None:
            return
        super(RepertoireGrid, self).on_data_change(instance)

    def set_selection_text(self):
        """Set status bar to display main PGN Tags."""
        if self.selection:
            ss0 = self.selection[0]
            if ss0 in self.objects:
                tags = self.objects[ss0].value.collected_game[1]
                self.ui.statusbar.set_status_text(
                    '  '.join(
                        [tags.get(k, '')
                         for k in REPERTOIRE_TAG_ORDER] +
                        [' '.join(
                            [t.group()
                             for t in self.objects[ss0].value.collected_game[2]
                             if t])]))
        else:
            self.ui.statusbar.set_status_text('')

    def launch_delete_record(self, key, modal=True):
        """Create delete dialogue."""
        oldobject = ChessDBrecordRepertoireUpdate()
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue))
        self.create_delete_dialog(
            self.objects[key],
            oldobject,
            modal,
            title='Delete Repertoire')

    def launch_edit_record(self, key, modal=True):
        """Create edit dialogue."""
        self.create_edit_dialog(
            self.objects[key],
            ChessDBrecordRepertoireUpdate(),
            ChessDBrecordRepertoireUpdate(),
            False,
            modal,
            title='Edit Repertoire')

    def launch_edit_show_record(self, key, modal=True):
        """Create edit dialogue including reference copy of original."""
        self.create_edit_dialog(
            self.objects[key],
            ChessDBrecordRepertoireUpdate(),
            ChessDBrecordRepertoireUpdate(),
            True,
            modal,
            title='Edit Repertoire')

    def launch_insert_new_record(self, modal=True):
        """Create insert dialogue."""
        newobject = ChessDBrecordRepertoireUpdate()
        instance = self.datasource.new_row()
        instance.srvalue = repr(EMPTY_REPERTOIRE_GAME + UNKNOWN_RESULT)
        self.create_edit_dialog(
            instance,
            newobject,
            None,
            False,
            modal,
            title='New Repertoire')

    def launch_show_record(self, key, modal=True):
        """Create show dialogue."""
        oldobject = ChessDBrecordRepertoireUpdate()
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue))
        self.create_show_dialog(
            self.objects[key],
            oldobject,
            modal,
            title='Show Repertoire')

    def make_display_widget(self, sourceobject):
        """Return a DatabaseRepertoireDisplay for sourceobject."""
        game = DatabaseRepertoireDisplay(
            master=self.ui.view_repertoires_pw,
            ui=self.ui,
            items_manager=self.ui.repertoire_items,
            itemgrid=self.ui.repertoire_games,
            sourceobject=sourceobject)
        game.set_position_analysis_data_source()
        game.pgn.get_first_game(sourceobject.get_srvalue())
        game.set_game()
        return game
        
    def make_edit_widget(self, sourceobject):
        """Return a DatabaseRepertoireEdit for sourceobject."""
        game = DatabaseRepertoireEdit(
            master=self.ui.view_repertoires_pw,
            ui=self.ui,
            items_manager=self.ui.repertoire_items,
            itemgrid=self.ui.repertoire_games,
            sourceobject=sourceobject)
        game.set_position_analysis_data_source()
        game.pgn.get_first_game(sourceobject.get_srvalue())
        game.set_game(reset_undo=True)
        return game

    def display_selected_item(self, key):
        """Display and return a DatabaseRepertoireDisplay for selected game."""
        selected = self.get_visible_record(key)
        if selected is None:
            return None
        # Should the Frame containing board and score be created here and
        # passed to DatabaseGameDisplay. (Which needs 'import Tkinter' above.)
        # Rather than passing the container where the Frame created by
        # DatabaseGameDisplay is to be put.
        # Yes because DatabaseGameEdit (see edit_selected_item) includes
        # extra widgets. Want to say game.widget.destroy() eventually.
        # Read make_display_widget for DatabaseGameDisplay and DatabaseGameEdit.
        game = self.make_display_widget(selected)
        self.ui.add_repertoire_to_display(game)
        self.ui.repertoire_items.increment_object_count(key)
        self.ui.repertoire_items.set_itemmap(game, key)
        self.set_properties(key)
        return game
        
    def edit_selected_item(self, key):
        """Display and return a DatabaseRepertoireEdit for selected game."""
        selected = self.get_visible_record(key)
        if selected is None:
            return None
        # Should the Frame containing board and score be created here and
        # passed to DatabaseGameDisplay. (Which needs 'import Tkinter' above.)
        # Rather than passing the container where the Frame created by
        # DatabaseGameEdit is to be put.
        # Yes because DatabaseGameDisplay (see display_selected_item) includes
        # less widgets. Want to say game.widget.destroy() eventually.
        # Read make_edit_widget for DatabaseGameDisplay and DatabaseGameEdit.
        game = self.make_edit_widget(selected)
        self.ui.add_repertoire_to_display(game)
        self.ui.repertoire_items.increment_object_count(key)
        self.ui.repertoire_items.set_itemmap(game, key)
        self.set_properties(key)
        return game

    def export_pgn(self, event=None):
        """Export selected repertoires like PGN."""
        exporters.export_grid_repertoires_as_pgn(
            self,
            self.ui.get_export_filename('Repertoires', pgn=True))

    def export_rav_pgn(self, event=None):
        """Export selected repertoires like PGN archive."""
        exporters.export_grid_repertoires_as_rav_pgn(
            self,
            self.ui.get_export_filename('RAV Repertoires', pgn=True))

    def is_visible(self):
        """Return True if list of repertoire games is displayed."""
        return str(self.get_frame()) in self.ui.repertoires_pw.panes()

    def set_properties(self, key, dodefaultaction=True):
        """Return True if properties for game key set or False."""
        if super(GameListGrid, self).set_properties(
            key, dodefaultaction=False):
            return True
        if self.ui.repertoire_items.object_display_count(key):
            self.objects[key].set_background_on_display(
                self.get_row_widgets(key))
            self.set_row_under_pointer_background(key)
            return True
        if dodefaultaction:
            self.objects[key].set_background_normal(self.get_row_widgets(key))
            self.set_row_under_pointer_background(key)
            return True
        return False

    def set_row(self, key, dodefaultaction=True, **kargs):
        """Return row widget for game key or None."""
        row = super(GameListGrid, self).set_row(
            key, dodefaultaction=False, **kargs)
        if row is not None:
            return row
        if key not in self.keys:
            return None
        if self.ui.repertoire_items.object_display_count(key):
            return self.objects[key].grid_row_on_display(**kargs)
        if dodefaultaction:
            return self.objects[key].grid_row_normal(**kargs)
        else:
            return None

    def focus_set_frame(self, event=None):
        """Delegate to superclass then set toolbar widget states."""
        super().focus_set_frame(event=event)
        self.ui.set_toolbarframe_normal(
            self.ui.move_to_repertoire, self.ui.filter_repertoire)

    def set_selection(self, key):
        """Hack to fix edge case when inserting records using apsw or sqlite3.
        
        Workaround a KeyError exception when a record is inserted while a grid
        keyed by a secondary index with only one key value in the index is on
        display.
        
        """
        try:
            super().set_selection(key)
        except KeyError:
            tkinter.messagebox.showinfo(
                title='Insert Repertoire Workaround',
                message=''.join(
                    ('All records have same name on this display.\n\nThe new ',
                     'record has been inserted but you need to Hide, and then ',
                     'Show, the display to see the record in the list.',
                     )))


class RepertoirePositionGames(GameListGrid):

    """Customized GameListGrid for list of games matching a repertoire position.

    The grid is populated by a FullPositionDS instance from the
    dpt.fullpositionds or basecore.fullpositionds modules.
    """

    def __init__(self, ui):
        """Extend with position grid definition and bindings.

        ui - container for user interface widgets and methods.

        """
        super(RepertoirePositionGames, self).__init__(
            ui.position_repertoires_pw, ui)
        self.make_header(ChessDBrowPosition.header_specification)
        self.__bind_on()
        for function, accelerator in (
            (self.display_game_from_popup,
             EventSpec.display_game_from_repertoire_game_grid),
            (self.edit_game_from_popup,
             EventSpec.edit_game_from_repertoire_game_grid),
            (self.export_archive_pgn,
             EventSpec.export_archive_pgn_from_repertoire_game_grid),
            (self.export_rav_pgn,
             EventSpec.export_rav_pgn_from_repertoire_game_grid),
            (self.export_pgn,
             EventSpec.export_pgn_from_repertoire_game_grid),
            (self.up_one_transposition,
             EventSpec.cycle_up_one_repeat_in_repertoire_game_grid),
            (self.down_one_transposition,
             EventSpec.cycle_down_one_repeat_in_repertoire_game_grid),
            ):
            self.menupopup.add_command(
                label=accelerator[1],
                command=self.try_command(function, self.menupopup),
                accelerator=accelerator[2])
        self.add_navigation_to_popup()
        self.add_navigation_to_popup_no_row()
        for function, accelerator in (
            (self.set_focus_position_grid,
             EventSpec.repertoire_game_grid_to_position_grid),
            (self.set_focus_gamepanel_item_command,
             EventSpec.repertoire_game_grid_to_active_game),
            (self.set_focus_game_grid,
             EventSpec.repertoire_game_grid_to_game_grid),
            (self.set_focus_repertoire_grid,
             EventSpec.repertoire_game_grid_to_repertoire_grid),
            (self.set_focus_repertoirepanel_item_command,
             EventSpec.repertoire_game_grid_to_active_repertoire),
            (self.set_focus_partial_grid,
             EventSpec.repertoire_game_grid_to_partial_grid),
            (self.set_focus_partialpanel_item_command,
             EventSpec.repertoire_game_grid_to_active_partial),
            (self.set_focus_partial_game_grid,
             EventSpec.repertoire_game_grid_to_partial_game_grid),
            (self.set_focus_selection_rule_grid,
             EventSpec.repertoire_game_grid_to_selection_rule_grid),
            (self.set_focus_selectionpanel_item_command,
             EventSpec.repertoire_game_grid_to_active_selection_rule),
            (self.traverse_backward,
             EventSpec.tab_traverse_backward),
            (self.traverse_forward,
             EventSpec.tab_traverse_forward),
            ):
            for m in (self.menupopup_navigation,
                      self.menupopup_navigation_no_row):
                m.add_command(
                    label=accelerator[1],
                    command=self.try_command(function, m),
                    accelerator=accelerator[2])

    def bind_off(self):
        """Disable all bindings."""
        super(RepertoirePositionGames, self).bind_off()
        for sequence, function in (
            (EventSpec.repertoire_game_grid_to_repertoire_grid, ''),
            (EventSpec.repertoire_game_grid_to_active_repertoire, ''),
            (EventSpec.repertoire_game_grid_to_partial_grid, ''),
            (EventSpec.repertoire_game_grid_to_active_partial, ''),
            (EventSpec.repertoire_game_grid_to_partial_game_grid, ''),
            (EventSpec.repertoire_game_grid_to_position_grid, ''),
            (EventSpec.repertoire_game_grid_to_active_game,
             self.set_focus_gamepanel_item),
            (EventSpec.repertoire_game_grid_to_game_grid, ''),
            (EventSpec.repertoire_game_grid_to_selection_rule_grid, ''),
            (EventSpec.repertoire_game_grid_to_active_selection_rule, ''),
            (EventSpec.display_game_from_repertoire_game_grid, ''),
            (EventSpec.edit_game_from_repertoire_game_grid, ''),
            (EventSpec.cycle_up_one_repeat_in_repertoire_game_grid, ''),
            (EventSpec.cycle_down_one_repeat_in_repertoire_game_grid, ''),
            (EventSpec.export_archive_pgn_from_repertoire_game_grid, ''),
            (EventSpec.export_rav_pgn_from_repertoire_game_grid, ''),
            (EventSpec.export_pgn_from_repertoire_game_grid, ''),
            ):
            if function:
                function = self.try_event(function)
            self.frame.bind(sequence[0], function)

    def bind_on(self):
        """Enable all bindings."""
        super(RepertoirePositionGames, self).bind_on()
        self.__bind_on()

    def __bind_on(self):
        """Enable all bindings."""
        for sequence, function in (
            (EventSpec.repertoire_game_grid_to_repertoire_grid,
             self.set_focus_repertoire_grid),
            (EventSpec.repertoire_game_grid_to_active_repertoire,
             self.set_focus_repertoirepanel_item),
            (EventSpec.repertoire_game_grid_to_partial_grid,
             self.set_focus_partial_grid),
            (EventSpec.repertoire_game_grid_to_active_partial,
             self.set_focus_partialpanel_item),
            (EventSpec.repertoire_game_grid_to_partial_game_grid,
             self.set_focus_partial_game_grid),
            (EventSpec.repertoire_game_grid_to_position_grid,
             self.set_focus_position_grid),
            (EventSpec.repertoire_game_grid_to_active_game,
             self.set_focus_gamepanel_item),
            (EventSpec.repertoire_game_grid_to_game_grid,
             self.set_focus_game_grid),
            (EventSpec.repertoire_game_grid_to_selection_rule_grid,
             self.set_focus_selection_rule_grid),
            (EventSpec.repertoire_game_grid_to_active_selection_rule,
             self.set_focus_selectionpanel_item),
            (EventSpec.display_game_from_repertoire_game_grid,
             self.display_game),
            (EventSpec.edit_game_from_repertoire_game_grid,
             self.edit_game),
            (EventSpec.cycle_up_one_repeat_in_repertoire_game_grid,
             self.up_one_transposition),
            (EventSpec.cycle_down_one_repeat_in_repertoire_game_grid,
             self.down_one_transposition),
            (EventSpec.export_archive_pgn_from_repertoire_game_grid,
             self.export_archive_pgn),
            (EventSpec.export_rav_pgn_from_repertoire_game_grid,
             self.export_rav_pgn),
            (EventSpec.export_pgn_from_repertoire_game_grid,
             self.export_pgn),
            ):
            if function:
                function = self.try_event(function)
            self.frame.bind(sequence[0], function)

    def display_game(self, event=None):
        """Display selected game and cancel selection."""
        self.set_move_highlight(
            self.display_selected_item(self.get_visible_selected_key()))
        self.cancel_selection()

    def display_game_from_popup(self, event=None):
        """Display game selected by pointer."""
        self.set_move_highlight(
            self.display_selected_item(self.pointer_popup_selection))

    def edit_game(self, event=None):
        """Display selected game with editing allowed and cancel selection."""
        self.set_move_highlight(
            self.edit_selected_item(self.get_visible_selected_key()))
        self.cancel_selection()

    def edit_game_from_popup(self, event=None):
        """Display game with editing allowed selected by pointer."""
        self.set_move_highlight(
            self.edit_selected_item(self.pointer_popup_selection))

    def set_row(self, key, dodefaultaction=True, **kargs):
        """Return row widget for game key or None.

        Add arguments to **kargs for grid_row method in PositionRow class.
        
        """
        kargs.update(
            position=self.datasource.fullposition,
            context=self.ui.get_active_repertoire_move())
        return super(RepertoirePositionGames, self).set_row(
            key, dodefaultaction=dodefaultaction, **kargs)

    def on_game_change(self, instance):
        # datasource refers to a set derived from file and may need
        # to be recreated
        if self.get_data_source() is None:
            return
        super(RepertoirePositionGames, self).on_data_change(instance)

    def down_one_transposition(self, event=None): # Hack for menu item
        """Cycle forward one tranposition to position in selected game."""
        if self.selection:
            ss0 = self.selection[0]
            if ss0 in self.objects:
                self.objects[ss0].score.down_one_transposition()
        
    def up_one_transposition(self, event=None): # Hack for menu item
        """Cycle backward one tranposition to position in selected game."""
        if self.selection:
            ss0 = self.selection[0]
            if ss0 in self.objects:
                self.objects[ss0].score.up_one_transposition()
        
    def set_selection_text(self):
        """Set status bar to display main PGN Tags."""
        if self.selection:
            ss0 = self.selection[0]
            if ss0 in self.objects:
                t = self.objects[ss0].value.collected_game[1]
                self.ui.statusbar.set_status_text(
                    '  '.join([t.get(k, '')
                               for k in STATUS_SEVEN_TAG_ROSTER_EVENT]))
        else:
            self.ui.statusbar.set_status_text('')

    def is_visible(self):
        """Return True if list of matching repertoire games is displayed."""
        return str(self.get_frame()) in self.ui.position_repertoires_pw.panes()

    def is_payload_available(self):
        """Return True if connected to database and games displayed."""
        if not super().is_payload_available():
            return False
        return self.ui.repertoire_items.is_visible()

    def make_display_widget(self, sourceobject):
        """Return a DatabaseGameDisplay for sourceobject."""
        game = super().make_display_widget(sourceobject)
        game.set_game()
        return game
        
    def make_edit_widget(self, sourceobject):
        """Return a DatabaseGameEdit for sourceobject."""
        game = super().make_edit_widget(sourceobject)
        game.set_game(reset_undo=True)
        return game

    def focus_set_frame(self, event=None):
        """Delegate to superclass then set toolbar widget states."""
        super().focus_set_frame(event=event)
        self.ui.set_toolbarframe_disabled()
